#include        <ctype.h>
#include        <string.h>
#include        "mes.h"
#include        "conf.h"
#include	"usblan.h"

BUSConf	sram_config;
BUSConf	dram_config;
BUSConf	flush_config;
BUSConf	memory_config;
BUSConf	usb_config[SL811_MAX];
BUSConf	ramdisk_config[RAMDISK_MAX];
BUSConf	romdisk_config[ROMDISK_MAX];
BUSConf	sci_config[SCI_MAX];
BUSConf	ne_config[NE_MAX];
BUSConf	wi_config[WI_MAX];
BUSConf	hdd_config[HDD_MAX];
PIOConf	lcd_config[LCD_MAX];
PIOConf	mmc_config[MMC_MAX];
PIOConf	i2c_config[I2C_MAX];
PIOConf	rtc_config[RTC_MAX];
char	master_devname[16], shell_name[16], stdio_device[16];
char	arp_setip, udp_netbios, timer_sw_flag, regsize, oprate, files;

#define		BUFSIZ		1024

#define		MONO_DEV	0
#define		NONE_TYPE	1
#define		BUS_TYPE	1
#define		PIO_TYPE	2

#define		SET_DATA	1
#define		AND_DATA	2
#define		OR_DATA		3

typedef struct {
	int	maxdev;
	int	type;
	char	devname[8];
	void	*confptr;
} DevInfo;

static const DevInfo dev_info[] = {
	{MONO_DEV,	NONE_TYPE, "none", 0},
	{MONO_DEV,	BUS_TYPE, "sram", &sram_config},
	{MONO_DEV,	BUS_TYPE, "dram", &dram_config},
	{MONO_DEV,	BUS_TYPE, "flush", &flush_config},
	{MONO_DEV,	BUS_TYPE, "memory", &memory_config},
	{SL811_MAX,	BUS_TYPE, "usb0", usb_config},
	{RAMDISK_MAX,	BUS_TYPE, "ram0", ramdisk_config},
	{ROMDISK_MAX,	BUS_TYPE, "rom0", romdisk_config},
	{SCI_MAX,	BUS_TYPE, "sci0", sci_config},
	{MMC_MAX,	PIO_TYPE, "mmc0", mmc_config},
	{LCD_MAX,	PIO_TYPE, "lcd0", lcd_config},
	{NE_MAX,	BUS_TYPE, "ne0", ne_config},
	{WI_MAX,	BUS_TYPE, "wi0", wi_config},
	{HDD_MAX,	BUS_TYPE, "hdd0", hdd_config},
	{I2C_MAX,	PIO_TYPE, "i2c0", i2c_config},
	{RTC_MAX,	PIO_TYPE, "rtc0", rtc_config},
	{MONO_DEV,	NONE_TYPE, "boot", 0},
	{MONO_DEV,	NONE_TYPE, "", 0},
};

enum device_type {
	none_type,
	sram_type,
	dram_type,
	flush_type,
	memory_type,
	usb_type,
	ramdisk_type,
	romdisk_type,
	sci_type,
	mmc_type,
	lcd_type,
	ne_type,
	wi_type,
	hdd_type,
	i2c_type,
	rtc_type,
	boot_type,
};

static int zeroconfig(void);
static int setconfig(char *word);
static int set_register(char *word);

static int zeroconfig(void) {
	int		b, n, minor, max;
	BUSConf		*bus_config;
	PIOConf		*pio_config;

	for(n = 1;dev_info[n].devname[0] != 0;n++) {
		if(dev_info[n].type == BUS_TYPE) {
			bus_config = dev_info[n].confptr;
			max = dev_info[n].maxdev;
			if(dev_info[n].maxdev == MONO_DEV) max = 1;
			for(minor = 0;minor < max;minor++) {
				bus_config->base0 = 0;
				bus_config->base1 = 0;
				bus_config->width = 8;
				bus_config->irq = 255;
				bus_config->res_port = 0xff;
				bus_config->res_bit = 0;
				bus_config->wait = 63;
				bus_config = &bus_config[1];
			}
		} else if(dev_info[n].type == PIO_TYPE) {
			pio_config = dev_info[n].confptr;
			max = dev_info[n].maxdev;
			if(dev_info[n].maxdev == MONO_DEV) max = 1;
			for(minor = 0;minor < max;minor++) {
				pio_config->port = 0;
				for(b = 0;b < 8;b++) pio_config->bits[b] = 0;
				pio_config->bits[5] = 0xff;
				pio_config = &pio_config[1];
			}
		}
	}
}

static volatile int msiz, madr;

static int setconfig(char *word) {
	static int	current = none_type, minor = 0;
	unsigned int	n, len;
	BUSConf		*bus_config;
	PIOConf		*pio_config;

	if(memcmp(word, "end", 3) == 0) return 0;
	if(memcmp(word, "arpsetip", 8) == 0) {
		arp_setip = 1;
		return 1;
	}
	if(memcmp(word, "netbios", 7) == 0) {
		udp_netbios = 1;
		return 1;
	}
	if(memcmp(word, "use_timer", 9) == 0) {
		timer_sw_flag = word[9] - '0';
		return 1;
	}
	if(memcmp(word, "files=", 6) == 0) {
		files = stoi(&word[6]);
		return 1;
	}
	for(n = 1;dev_info[n].devname[0] != 0;n++) {
		len = strlen(dev_info[n].devname);
		if(isdigit(dev_info[n].devname[len - 1])) {
			if(memcmp(word, dev_info[n].devname, len - 1) == 0) {
				len = strlen(word);
				minor = word[len - 1] - '0';
				if(minor >= dev_info[n].maxdev) {
					current = none_type;
				} else {
					current = n;
				}
				return 1;
			}
		} else {
			if(strcmp(word, dev_info[n].devname) == 0) {
				current = n;
				return 1;
			}
		}
	}
	if(dev_info[current].type == BUS_TYPE) {
		bus_config = dev_info[current].confptr;
		if(dev_info[current].maxdev != MONO_DEV) {
			bus_config = &bus_config[minor];
		}
		if(memcmp(word, "b=", 2) == 0) {
			bus_config->memaddr = stoh(&word[2]);
			madr = bus_config->memaddr;
		} else if(memcmp(word, "s=", 2) == 0) {
			bus_config->memsize = stoh(&word[2]);
			msiz = bus_config->memaddr;
		} else if(memcmp(word, "ar=", 3) == 0) {
			bus_config->addr_reg = stoh(&word[3]);
		} else if(memcmp(word, "dr=", 3) == 0) {
			bus_config->data_reg = stoh(&word[3]);
		} else if(memcmp(word, "f=", 2) == 0) {
			bus_config->scifreq = stoi(&word[2]);
		} else if(memcmp(word, "br=", 3) == 0) {
			bus_config->scispeed = stoi(&word[3]);
		} else if(memcmp(word, "c=", 2) == 0) {
			bus_config->width = 16;
			bus_config->base1 = stoh(&word[2]);
		} else if(memcmp(word, "irq=", 4) == 0) {
			bus_config->irq = stoi(&word[4]);
		} else if(memcmp(word, "bw=", 3) == 0) {
			bus_config->width = stoi(&word[3]);
		} else if(memcmp(word, "w=", 2) == 0) {
			bus_config->wait = stoi(&word[2]);
		} else if(memcmp(word, "port=", 5) == 0) {
			bus_config->res_port = (stoh(&word[5]) >> 4) & 0x0f;
			bus_config->res_bit = stoh(&word[5]) & 0x0f;
		}
	}
	if(dev_info[current].type == PIO_TYPE) {
		pio_config = dev_info[current].confptr;
		pio_config = &pio_config[minor];
		if(memcmp(word, "p=", 2) == 0) {
			pio_config->port = stoi(&word[2]);
		} else if(memcmp(word, "cs=", 3) == 0) {
			pio_config->bits[0] = stoi(&word[3]);
		} else if(memcmp(word, "sio=", 4) == 0) {
			pio_config->bits[1] = stoi(&word[4]);
		} else if(memcmp(word, "clk=", 4) == 0) {
			pio_config->bits[1] = stoi(&word[4]);
		} else if(memcmp(word, "cmd=", 4) == 0) {
			pio_config->bits[2] = stoi(&word[4]);
		} else if(memcmp(word, "dat=", 4) == 0) {
			pio_config->bits[3] = stoi(&word[4]);
		} else if(memcmp(word, "csp=", 4) == 0) {
			pio_config->bits[4] = stoi(&word[4]);
		} else if(memcmp(word, "spd=", 4) == 0) {
			pio_config->bits[7] = stoi(&word[4]);
		} else if(memcmp(word, "eject=", 6) == 0) {
			pio_config->bits[5] = (stoh(&word[6]) >> 4) & 0x0f;
			pio_config->bits[6] = stoh(&word[6]) & 0x0f;
		} else if(memcmp(word, "d4=", 3) == 0) {
			pio_config->bits[0] = stoi(&word[3]);
		} else if(memcmp(word, "d5=", 3) == 0) {
			pio_config->bits[1] = stoi(&word[3]);
		} else if(memcmp(word, "d6=", 3) == 0) {
			pio_config->bits[2] = stoi(&word[3]);
		} else if(memcmp(word, "d7=", 3) == 0) {
			pio_config->bits[3] = stoi(&word[3]);
		} else if(memcmp(word, "rs=", 3) == 0) {
			pio_config->bits[4] = stoi(&word[3]);
		} else if(memcmp(word, "e=", 2) == 0) {
			pio_config->bits[5] = stoi(&word[2]);
		} else if(memcmp(word, "row=", 4) == 0) {
			pio_config->bits[6] = stoi(&word[4]);
		} else if(memcmp(word, "col=", 4) == 0) {
			pio_config->bits[7] = stoi(&word[4]);
		} else if(memcmp(word, "size=", 5) == 0) {
			pio_config->size = stoh(&word[5]);
		}
	}
	if(current == boot_type) {
		if(memcmp(word, "dev=", 4) == 0) {
			strncpy(master_devname, &word[4], 15);
			master_devname[15] = 0;
		} else if(memcmp(word, "shell=", 6) == 0) {
			strncpy(master_devname, &word[6], 15);
			shell_name[15] = 0;
		} else if(memcmp(word, "stdio=", 6) == 0) {
			strncpy(stdio_device, &word[6], 15);
			stdio_device[15] = 0;
		}
	}
	return 1;
}

void config(char *src, char *dst, int size) {
	char	*ptr;
	int	ret;

	master_devname[0] = 0;
	shell_name[0] = 0;
	stdio_device[0] = 0;
	arp_setip = 0;
	udp_netbios = 0;
	files = 8;
	timer_sw_flag = -1;
	zeroconfig();
	memcpy(dst, src, size - 1);
	dst[size] = 0;
	ptr = strtok(dst, " \n\r");
	ret = setconfig(ptr);
	while(ptr != 0 && ret == 1) {
		ptr = strtok(0," \n\r");
		if(ptr != 0) ret = setconfig(ptr);
	}
}

static int set_register(char *word) {
	char		*ptr;
	int		addr, value, i, n;
	volatile char	*byteword;
	volatile short	*halfword;
	volatile int	*longword;

	if(memcmp(word, "end", 3) == 0) return 0;
	if(memcmp(word, "byte", 4) == 0) {
		regsize = 1;
	} else if(memcmp(word, "half", 4) == 0) {
		regsize = 2;
	} else if(memcmp(word, "long", 4) == 0) {
		regsize = 4;
	} else if(memcmp(word, "set", 3) == 0) {
		oprate = SET_DATA;
	} else if(memcmp(word, "and", 3) == 0) {
		oprate = AND_DATA;
	} else if(memcmp(word, "or", 2) == 0) {
		oprate = OR_DATA;
	} else {
		n = strlen(word);
		if(word[0] == '=') return 1;
		for(i = 0;i < n;i++) {
			if(isxdigit(word[i])) continue;
			if(word[i] == '=') continue;
			return 1;
		}
		ptr = strchr(word, '=');
		if(ptr == 0) return 1;
		*ptr = 0;
		ptr++;
		addr = stoh(word);
		value = stoh(ptr);
		switch(regsize) {
		case 1:
			byteword = (volatile char *)addr;
			switch(oprate) {
			case SET_DATA:
				*byteword = value;
				break;
			case AND_DATA:
				*byteword &= ~value;
				break;
			case OR_DATA:
				*byteword |= value;
				break;
			}
			break;
		case 2:
			halfword = (volatile short *)addr;
			switch(oprate) {
			case SET_DATA:
				*halfword = value;
				break;
			case AND_DATA:
				*halfword &= ~value;
				break;
			case OR_DATA:
				*halfword |= value;
				break;
			}
			break;
		case 4:
			longword = (volatile int *)addr;
			switch(oprate) {
			case SET_DATA:
				*longword = value;
				break;
			case AND_DATA:
				*longword &= ~value;
				break;
			case OR_DATA:
				*longword |= value;
				break;
			}
			break;
		}
	}
	return 1;	
}

void init_register(char *src, char *dst, int size) {
	char	*ptr;
	int	ret;

#if defined(__H8300H__)
	regsize = 1;
#endif
#if defined(__sh__)
	regsize = 2;
#endif
	oprate = SET_DATA;
	memcpy(dst, src, size - 1);
	dst[size] = 0;
	ptr = strtok(dst, " \n\r");
	ret = set_register(ptr);
	while(ptr != 0 && ret == 1) {
		ptr = strtok(0," \n\r");
		if(ptr != 0) ret = set_register(ptr);
	}
}
