/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/2/14 first release		*/
/*					*/
/****************************************/
#include "mes.h"
#include "conf.h"
#include <regdef.h>

extern BUSConf	sram_config;
extern BUSConf	dram_config;
extern BUSConf	flush_config;

int init_stage1(void) {
	char	*conf;
	int	len, ret;

#if defined(__arm__)
	EBI_CSR0 = 0x400000 + EBI_DBW_16 + EBI_NWS_8 + EBI_WSE + EBI_PAGES_1M + EBI_TDF_7 + EBI_CSEN;
	EBI_CSR1 = 0x600000 + EBI_DBW_8 + EBI_NWS_2 + EBI_WSE + EBI_PAGES_1M + EBI_TDF_0 + EBI_CSEN;
	EBI_CSR2 = 0x800000 + EBI_DBW_16 + EBI_NWS_4 + EBI_WSE + EBI_PAGES_1M + EBI_TDF_7 + EBI_CSEN;
	EBI_CSR3 = 0xa00000 + EBI_DBW_16 + EBI_NWS_4 + EBI_WSE + EBI_PAGES_1M + EBI_TDF_7 + EBI_CSEN;
	PIO_PER = P25;
	PIO_OER = P25;
	PIO_CODR = P25;
#endif
#if defined(__H8300H__)
	ddr_set(1, 0xff);
	ddr_set(2, 0xff);
	ddr_set(3, 0xff);
	ddr_reset(4, 0xff);
	ddr_set(5, 0xff);
	ddr_reset(6, 0xff);
	ddr_set(8, 0xff);
	ddr_reset(9, 0xff);
	ddr_reset(10, 0xff);
	ddr_reset(11, 0xff);
	SYSCR = SYSCR & ~0x08;
	IPRA = 0x07;	// ITU0-2
	IPRB = 0xc0;	// ITU3-4
  #if defined(H8_3052)
	MSTCR |= 0x80;
  #endif
  #if defined(H8_3069)
	MSTCRH |= 0x80;
  #endif
#endif
#if defined(SH_7045)
	SYSCR = 0xc3;		/* AUD Reset disabled, internal ram enabled */
	MSTCR1 = 0xf030;	/* Standby disabled */
	MSTCR2 = 0xc0f0;
	BCR2 = 0xf7dd;		/* wait 1 idle */
	PACRH = 0x5004;		/* WRHH WRHL -WAIT are enable */
	PACRL1 = 0x155a;	/* -RD -WRH -WRL -CS[1..0] -IRQ[2..3] are enable */
	PACRL2 = 0xa155;	/* -CS[3..2] TXD1 RXD1 SCK0 TXD0 RXD0 are enable */
	PBCR1 = 0x000a;		/* A21/A20 are enable */
	PBCR2 = 0xafff;		/* -RDWR -CASH -CASL -RAS A19/A18/A17/16 are enable */
	PCCR = 0xffff;		/* A15-0 are enable */
	PDCRH1 = 0x0000;	/* D31-24 are disable */
	PDCRH2 = 0x0000;	/* D23-16 are disable */
	PECR1 = 0x0000;
	PECR2 = 0x0000;
	PECR1 = 0x0022;		/* SCK2 RXD2 are enable */
	PECR2 = 0xaa00;		/* TXD2 SCK3 TXD3 RXD3 are enable */
	PDCRL = 0xffff;		/* D[15..0] are enable */	
	PDCRL2 = 0x0000;	/* D[15..0] are enable */
#endif
#if defined(SH_7706)
	ICR1 = 0x0000;		/* IRQ Interrupt */
	PHCR = 0x2aaa;		/* IRQ Disable */
	IRR0 = 0x0000;
	IPRB |= 0xf000;		/* WDT */
#endif
	conf = get_romconfig(&len, "config.sys");
	if(conf != 0) {
		char	buf[256];

		config(conf, buf, 256);
	}
  #if defined(H8_3069)
	if(dram_config.memsize > 0) {
		set_buswidth(dram_config.memaddr, dram_config.width, dram_config.wait);
		RTCOR = 10;
		RTMCSR = 0x18;
		DRCRB = 0x98;
		DRCRA = 0x30;
	}
  #endif
  #if defined(SH_7045)
	if(dram_config.memsize > 0) {
		set_buswidth(dram_config.memaddr, dram_config.width, dram_config.wait);
		WCR2 = 0x000f;		/* default */
		DCR = 0x0021;		/* 8bit-bus 10bit-address */
		RTCSR = 0x001a;		/* clk x 2048 */
		RTCOR = 10;		/* 1.5[ms] reflesh cycle */
	}
  #endif
	if(sram_config.memsize > 0) {
		set_buswidth(sram_config.memaddr, sram_config.width, sram_config.wait);
	}
	if(flush_config.memsize > 0) {
		set_buswidth(flush_config.memaddr, flush_config.width, flush_config.wait);
	}
	ret = mm_init();
	init_regist();
	return ret;
}

void init_stage2(void) {
	char	*conf;
	int	len;

	conf = get_romconfig(&len, "init.sys");
	if(conf != 0) {
		char	buf[len];

		init_register(conf, buf, len);
	}
	conf = get_romconfig(&len, "config.sys");
	if(conf != 0) {
		char	buf[len];

		config(conf, buf, len);
	}
	fat_init();
	interrupt_init();
	shm_init();
	device_init();
	usb_init();
	load_disk();
	ip_init();
}
