/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/2/14 first release		*/
/*					*/
/****************************************/
#include <macro.h>
#include <string.h>
#include "mes.h"
#include "id.h"
#include "ethernet.h"
#include "arp.h"
#include "ip.h"
#include "icmp.h"
#include "udp.h"
#include "tcp.h"

#define NETNUM	3
#define CACHESIZ 128

#define	IPDATA_KEEP	20000

#define READ_ACCESS	1
#define WRITE_ACCESS	2
#define NO_ACCESS	0

typedef volatile struct {
	int		fd;
	unsigned int	ip;
	unsigned int	mask;
	unsigned int	dhcp;
	Packet		arp, packet;
	Packet		que;
	unsigned short	ip_id;
	unsigned short	icmp_id;
	char		mac[ETHER_ADDR_LEN];
	char		devname[9];
	char		access;
} NetInfo;

extern char arp_setip;
extern char udp_netbios;
int tcp_recv(int net, IPHdr *iphdr);

static volatile unsigned int	gateway, gw_net, dns;
static NetInfo			*netinfo;
static volatile unsigned char	**macs;
static volatile unsigned int	*ips;
static char			*arpbuf, *netdatbuf;

static int	ip2mac(unsigned int ip, char *mac);
static void	icmp(int net, IPHdr *iphdr);
static void	ip_recv(int net, Packet *packet);
static void	arp(int net, char *data);
//static int	check_ip(int addr);
static void	net_data(int net);
static void	send_ip(int net, Packet *packet, int flag);
static int	set_cache(char *mac, unsigned int ip);

short checksum(char *data, int size, int init) {
	int	i, sum, mask;
	short	*ptr;

	sum = init & 0xffff;
	ptr = (short*)data;
	for(i = 0;i < ((size + 1) / 2);i++) {
		if(i == (size / 2) && (size & 1)) mask = 0xff00;
		else mask = 0xffff;
		sum += (int)be16toh(*ptr) & mask;
		if(sum >= 0x10000) {
			sum &= 0xffff;
			sum++;
		}
		ptr++;
	}
	return ~sum;
}

int get_net(int ip) {
	int	net, netip;

	for(net = 0;net < NETNUM;net++) {		
		if(netinfo[net].fd != 0) {
			netip = netinfo[net].ip & netinfo[net].mask;
			if((ip & netinfo[net].mask) == netip) return net;
		}
	}
	if(gateway != 0) return gw_net;
	return -1;
}

int get_ipinfo(int net, int *ip, int *mask, char *mac){
	if(net >= NETNUM) return -1;
	if(netinfo[net].fd == 0) return -1;
	memcpy(mac, (char*)netinfo[net].mac, ETHER_ADDR_LEN);
	*ip = netinfo[net].ip;
	*mask = netinfo[net].mask;
	return NETNUM;
}

int dhcp_server(int net, int serv) {
	if(net >= NETNUM) return -1;
	if(netinfo[net].fd == 0) return -1;
	if(serv != 0) netinfo[net].dhcp = serv;
	return netinfo[net].dhcp;
}

int setip(int net, int ip, int mask) {
	if(net >= NETNUM) return -1;
	if(netinfo[net].fd == 0) return -1;
	netinfo[net].ip = ip;
	netinfo[net].mask = mask;
	return 0;
}

static int set_cache(char *mac, unsigned int ip) {
	int		i, net;
	unsigned int	netip;

	if(mac == 0) {
		for(i = 0;i < CACHESIZ;i++) ips[i] = 0;
		return 0;
	}
	for(net = 0;net < NETNUM;net++) {		
		if(netinfo[net].fd != 0) {
			if(netinfo[net].ip == ip) return 0;
			netip = netinfo[net].ip & netinfo[net].mask;
			if((ip & netinfo[net].mask) == netip) {
				if((ip | netinfo[net].mask) == 0xffffffff) return 0;
				for(i = 0;i < CACHESIZ;i++) {
					if(ips[i] == ip) return 0;
					if(ips[i] == 0) {
						ips[i] = ip;
						memcpy((char*)macs[i], mac, ETHER_ADDR_LEN);
						return 0;
					}
				}
				for(i = 0;i < (CACHESIZ / 2);i++) ips[i] = ips[(CACHESIZ / 2) + i];
				for(;i < CACHESIZ;i++) ips[i] = 0;
				ips[CACHESIZ / 2] = ip;
				memcpy((char*)macs[CACHESIZ / 2], mac, ETHER_ADDR_LEN);
				return 0;
			}
		}
	}
	return -1;
}

static int ip2mac(unsigned int ip, char *mac) {
	int	i, net, netip;
	char	gw;

	if(ip == 0xffffffff) {
			memset(mac, 0xff, ETHER_ADDR_LEN);
			return 0;
	}
	gw = 1;
	for(net = 0;net < NETNUM;net++) {		
		if(netinfo[net].fd != 0) {
			if(netinfo[net].ip == ip) {
				memcpy(mac, (char*)netinfo[net].mac, ETHER_ADDR_LEN);
				return 0;
			}
			netip = netinfo[net].ip & netinfo[net].mask;
			if((ip & netinfo[net].mask) == netip) {
				gw = 0;
				if((ip | netinfo[net].mask) == 0xffffffff) {
					memset(mac, 0xff, ETHER_ADDR_LEN);
					return 0;
				}
			}
		}
	}
	for(i = 0;i < CACHESIZ;i++) {
		if(ips[i] == ip || (gw == 1 && ips[i] == gateway)) {
			memcpy(mac, (char*)macs[i], ETHER_ADDR_LEN);
			return 0;
		}
	}
	return -1;
}

void send_arp(int net, int ip) {
	char	*buf;
	ARPHdr	*arp;
	Ether	*ether;
	int	netip;

	netip = netinfo[net].ip & netinfo[net].mask;
	if((ip & netinfo[net].mask) != netip) {
		if(gateway == 0) return;
		ip = gateway;
	}
	buf = __malloc(SuperTask, ETHER_MIN_LEN);
	if(buf == 0) return;
	bzero(buf, ETHER_MIN_LEN);
	ether = (Ether*)buf;
	arp = (ARPHdr*)&buf[sizeof(Ether)];
	memset((char*)ether->dmac, 0xff, ETHER_ADDR_LEN);
	memcpy((char*)ether->smac, (char*)netinfo[net].mac, ETHER_ADDR_LEN);
	ether->type = htobe16(ETHERTYPE_ARP);
	arp->ar_hrd = htobe16(ARPHRD_ETHER);
	arp->ar_pro = htobe16(ETHERTYPE_IP);
	arp->ar_hln = 6;
	arp->ar_pln = 4;
	arp->ar_op = htobe16(ARPOP_REQUEST);
	bzero((char*)arp->ar_dmac, ETHER_ADDR_LEN);
	memcpy((char*)arp->ar_smac, (char*)netinfo[net].mac, ETHER_ADDR_LEN);
	arp->ar_dip = htobe32(ip);
	arp->ar_sip = htobe32(netinfo[net].ip);
	__write(SuperTask, netinfo[net].fd, buf, sizeof(Ether) + sizeof(ARPHdr));
	__free(SuperTask, buf);
}

static void send_ip(int net, Packet *packet, int flag) {
	Ether		*ether;
	IPHdr		*iphdr;
	int		addr, size, dip;

	ether = (Ether*)packet->data;
	addr = (int)packet->data + sizeof(Ether);
	iphdr = (IPHdr*)addr;
	dip = be32toh(iphdr->ip_dst);
	ip2mac(dip, (char*)ether->dmac);
	size = be16toh(iphdr->ip_len) + sizeof(Ether);
	if(flag == 0) {
		__write(SuperTask, netinfo[net].fd, (char*)packet->data, size);
		bzero((char*)packet->data, size);
		__free(MALLOC_IP, (char*)packet);
	} else {

		/************************************/
		/************* ARP TASK *************/
		/************************************/
		command_trap(_write_, netinfo[net].fd, (char*)packet->data, size);
		bzero((char*)packet->data, size);
		command_trap(_sysfree_, MALLOC_IP, (char*)packet);
	}
}

#define MTU 1480

void write_ip(int net, char *_data, int size, int proto, int src, int dst, int opt) {
	short	ip_id;
	int	ret, addr, iphdrsiz, sendsiz, sum, packsiz, remain, cursiz, off;
	char	*data;
	Ether	*ether;
	IPHdr	*iphdr;
	Packet	*packet, *cur;

	remain = size;
	++(netinfo[net].ip_id);
	do {
		data = &_data[size - remain];
		off = (size - remain) >> 3;
		cursiz = (remain > MTU) ? MTU : remain;
		remain -= cursiz;
		iphdrsiz = sizeof(IPHdr);
		sendsiz = sizeof(Ether) + iphdrsiz + cursiz;
		packsiz = sendsiz;
		packet = (Packet*)__malloc(MALLOC_IP, packsiz + sizeof(Packet));
		if(packet == 0) return;
		bzero((char*)packet, packsiz + sizeof(Packet));
		packet->stamp = task_time();
		packet->next = 0;
		bzero((char*)packet->data, sendsiz);
		ether = (Ether*)packet->data;
		addr = (int)packet->data + sizeof(Ether);
		iphdr = (IPHdr*)addr;
		iphdr->ip_ver_hl = (iphdrsiz >> 2) | 0x40;
		iphdr->ip_tos = 0;
		iphdr->ip_len = htobe16(cursiz + iphdrsiz);
		ip_id = netinfo[net].ip_id;
		iphdr->ip_id = htobe16(ip_id);
		if(remain != 0) off |= IP_MF;
		iphdr->ip_off = htobe16(off);
		iphdr->ip_ttl = 128;
		iphdr->ip_p = proto;
		iphdr->ip_sum = 0;
		iphdr->ip_src = htobe32(src);
		iphdr->ip_dst = htobe32(dst);
		sum = checksum((char*)iphdr, iphdrsiz, 0);
		iphdr->ip_sum = htobe16(sum);
		addr = (int)packet->data + sizeof(Ether) + iphdrsiz;
		memcpy((char*)addr, data, cursiz);
		memcpy((char*)ether->smac, (char*)netinfo[net].mac, ETHER_ADDR_LEN);
		ether->type = htobe16(ETHERTYPE_IP);
		ret = ip2mac(dst, (char*)ether->dmac);
		if(ret != -1) {
			send_ip(net, packet, 0);
		} else {
			task_sw_ctl(1);
			cur = &netinfo[net].que;
			while(cur->next != 0) cur = cur->next;
			cur->next = packet;
			task_sw_ctl(0);
			send_arp(net, dst);
		}
	} while(remain > 0);
}

static void icmp(int net, IPHdr *iphdr) {
	ICMP		*icmp;
	unsigned char	type;
	unsigned int	sip, dip;
	int		hsiz, icmpsiz, addr, sum, seq, task_id;

	hsiz = (iphdr->ip_ver_hl & 0x0f) << 2;
	sip = be32toh(iphdr->ip_src);
	dip = be32toh(iphdr->ip_dst);
	addr = (int)iphdr + hsiz;
	icmp = (ICMP*)addr;
	icmpsiz = be16toh(iphdr->ip_len) - hsiz;
	type = icmp->icmp_type;
	switch(type) {
	case ICMP_ECHO_REQUEST:
		icmp->icmp_type = ICMP_ECHO_REPLY;
		icmp->icmp_cksum = 0;
		sum = checksum((char*)icmp, icmpsiz, 0);
		icmp->icmp_cksum = htobe16(sum);
		write_ip(net, (char*)icmp, icmpsiz, ICMP_PROTOCOL, dip, sip, 0);
		break;
	case ICMP_ECHO_REPLY:
		seq = be16toh(icmp->icmp_seq);
		task_id = be16toh(icmp->icmp_id);
		set_arg(task_id, _ping, 6, seq);
		break;
	}
}

static void ip_recv(int net, Packet *packet) {
	Ether	*ether;
	IPHdr	*iphdr;
	int	i, hsiz;
	char	*data;
	unsigned int	sip, dip;
	unsigned short	cksum, src_sum, flag;

	data = (char*)(packet->data);
	ether = (Ether*)data;
	iphdr = (IPHdr*)(data + sizeof(Ether));
	flag = be16toh(iphdr->ip_off);
	if(flag & (IP_MF | IP_OFFMASK)) return;
	hsiz = (iphdr->ip_ver_hl & 0x0f) << 2;
	src_sum = be16toh(iphdr->ip_sum);
	iphdr->ip_sum = 0;
	cksum = checksum((char*)iphdr, hsiz, 0);
	if(src_sum != cksum) return;
	sip = be32toh(iphdr->ip_src);
	dip = be32toh(iphdr->ip_dst);
	if(arp_setip == 1 && netinfo[net].ip == 0) {
		netinfo[net].ip = dip;
		if((dip & 0xffffff00) == (sip & 0xffffff00)) {
			netinfo[net].mask = 0xffffff00;
		} else if((dip & 0xffff0000) == (sip & 0xffff0000)) {
			netinfo[net].mask = 0xffff0000;
		} else if((dip & 0xff000000) == (sip & 0xff000000)) {
			netinfo[net].mask = 0xff000000;
		}
	}
	switch(iphdr->ip_p) {
	case ICMP_PROTOCOL:
		icmp(net, iphdr);
		break;
	case UDP_PROTOCOL:
		udp_recv(net, packet);
		break;
	case TCP_PROTOCOL:
		for(i = 0;i < ETHER_ADDR_LEN;i++) {
			if(ether->dmac[i] != 0xff) break;
		}
		if(i < ETHER_ADDR_LEN) tcp_recv(net, iphdr);
		break;
	}
	if(iphdr->ip_p != UDP_PROTOCOL) __free(MALLOC_IP, (char*)packet);
}

/************************************/
/************* ARP TASK *************/
/************************************/
static void arp(int net, char *data) {
	Packet	*send, *cur;
	ARPHdr	*arp;
	IPHdr	*iphdr;
	Ether	*ether;
	char	*smac, *dmac;
	unsigned short proto, op;
	unsigned int	sip, dip, mydip;

	arp = (ARPHdr*)(data + sizeof(Ether)); 
	proto = be16toh(arp->ar_pro);
	if(proto != ETHERTYPE_IP) return;
	op = be16toh(arp->ar_op);
	sip = be32toh(arp->ar_sip);	
	dip = be32toh(arp->ar_dip);
	if(dip != netinfo[net].ip) return;
	smac = (char*)arp->ar_smac;
	dmac = (char*)arp->ar_dmac;
	set_cache((char*)arp->ar_smac, sip);
	switch(op) {
	case ARPOP_REQUEST:
		ether = (Ether*)arpbuf;
		arp = (ARPHdr*)&arpbuf[sizeof(Ether)];
		memcpy((char*)ether->dmac, smac, ETHER_ADDR_LEN);
		memcpy((char*)ether->smac, (char*)netinfo[net].mac, ETHER_ADDR_LEN);
		ether->type = htobe16(ETHERTYPE_ARP);
		arp->ar_hrd = htobe16(ARPHRD_ETHER);
		arp->ar_pro = htobe16(ETHERTYPE_IP);
		arp->ar_hln = 6;
		arp->ar_pln = 4;
		arp->ar_op = htobe16(ARPOP_REPLY);
		memcpy((char*)arp->ar_dmac, smac, ETHER_ADDR_LEN);
		memcpy((char*)arp->ar_smac, (char*)netinfo[net].mac, ETHER_ADDR_LEN);
		arp->ar_dip = htobe32(sip);
		arp->ar_sip = htobe32(netinfo[net].ip);
		command_trap(_write_, netinfo[net].fd, arpbuf, sizeof(Ether) + sizeof(ARPHdr));
		break;
	case ARPOP_REPLY:
		for(;;) {
			task_sw_ctl(1);
			send = 0;
			cur = &netinfo[net].que;
			while(cur->next != 0) {
				iphdr = (IPHdr*)(cur->next->data + sizeof(Ether)); 
				mydip = be32toh(iphdr->ip_dst);
				if(sip == mydip) {
					send = cur->next;
					cur->next = cur->next->next;
					break;
				}
				cur = cur->next;
			}
			task_sw_ctl(0);
			if(send == 0) break;
			send_ip(net, send, 1);
		}
		break;
	}
}

/*
static int check_ip(int addr) {
	IPHdr	*iphdr;
	char	mac[ETHER_ADDR_LEN];

	iphdr = (IPHdr*)addr;
	if(ip2mac(iphdr->ip_src, mac) == -1) return -1;
	if(ip2mac(iphdr->ip_dst, mac) == -1) return -1;
	return 0;
}
*/

/************************************/
/************* ARP TASK *************/
/************************************/
void arp_main(void) {
	Packet	*packet, *root;
	int	net;

	init_process();
	for(;;) {
		for(net = 0;net < NETNUM;net++) {
			if(netinfo[net].fd != 0) {
				net_data(net);
				for(;;) {
					task_sw_ctl(1);
					root = &netinfo[net].arp;
					if(root->next != 0) {
						packet = root->next;
						root->next = root->next->next;
					} else {
						task_sw_ctl(0);
						break;
					}
					task_sw_ctl(0);
					arp(net, (char*)packet->data);
					command_trap(_sysfree_, MALLOC_IP, (char*)packet);
				}
			}
		}
		usb_interval();
		task_sw_trap();
	}
}

/************************************/
/************* ARP TASK *************/
/************************************/
static void net_data(int net) {
	short	type;
	int	recvsize, dip, iphsiz;
	short	sport, dport;
	Packet	*cur, *packet;
	Ether	*ether;
	ARPHdr	*arp;
	IPHdr	*iphdr;
	UDPHdr	*udphdr;

	cur = 0;
	ether = (Ether*)netdatbuf;
	arp = (ARPHdr*)&netdatbuf[sizeof(Ether)];
	iphdr = (IPHdr*)&netdatbuf[sizeof(Ether)];
	for(;;) {
		recvsize = command_trap(_read_, netinfo[net].fd, netdatbuf, ETHER_MAX_LEN);
		if(recvsize <= 0) break;
		type = be16toh(ether->type);
		if(type == ETHERTYPE_ARP) {
			dip = be32toh(arp->ar_dip);
			if(dip != netinfo[net].ip) continue;
		} else if(type == ETHERTYPE_IP) {
			iphsiz = (iphdr->ip_ver_hl & 0x0f) << 2;
			if(iphdr->ip_p == ICMP_PROTOCOL || iphdr->ip_p == TCP_PROTOCOL) {
				dip = be32toh(iphdr->ip_dst);
				if(dip != netinfo[net].ip && netinfo[net].ip) continue;
			} else if(iphdr->ip_p == UDP_PROTOCOL) {
				if(udp_netbios == 0) {
					udphdr = (UDPHdr*)&netdatbuf[sizeof(Ether) + iphsiz];
					sport = be16toh(udphdr->dg_sport);
					dport = be16toh(udphdr->dg_dport);
					if(sport == 137 || sport == 138) continue;
					if(dport == 137 || dport == 138) continue;
				}
			} else {
				continue;
			}
			set_cache((char*)ether->smac, iphdr->ip_src);
		} else {
			continue;
		}
		packet = (Packet*)command_trap(_sysmalloc_, MALLOC_IP, recvsize + sizeof(Packet));
		if(packet == 0) return;
		packet->stamp = task_time();
		packet->next = 0;
		memcpy((char*)packet->data, netdatbuf, recvsize);
		task_sw_ctl(1);
		switch(type) {
		case ETHERTYPE_ARP:
			cur = &netinfo[net].arp;
			break;
		case ETHERTYPE_IP:
			cur = &netinfo[net].packet;
			break;
		}
		while(cur->next != 0) cur = cur->next;
		cur->next = packet;
		task_sw_ctl(0);
	}
}

int get_netnum(char* name) {
	int	net;

	for(net = 0;net < NETNUM;net++) {
		if(strcmp(name, (char*)netinfo[net].devname) == 0) {
			return net;
		}
	}
	return -1;
}

int net_ioctl(int net, int data, int op) {
	if(net >= NETNUM) return -1;
	return __ioctl(SuperTask, netinfo[net].fd, data, op);
}

int ifconfig(char* name, int ip, int mask) {
	int	i, net;

	for(net = 0;net < NETNUM;net++) {
		if(strcmp(name, (char*)netinfo[net].devname) == 0) {
			setip(net, ip, mask);
			return net;
		}
	}
	for(net = 0;net < NETNUM;net++) {
		if(netinfo[net].fd == 0) break;
	}
	if(net == NETNUM) return -1;
	netinfo[net].fd = __open(SuperTask, name, 0);
	if(netinfo[net].fd == -1) {
		netinfo[net].fd = 0;
		return -1;
	}
	for(i = 0;i < CACHESIZ;i++) ips[i] = 0;
	__ioctl(SuperTask, netinfo[net].fd, (int)netinfo[net].mac, GET_MAC);
	netinfo[net].ip = ip;
	netinfo[net].mask = mask;
	netinfo[net].dhcp = 0;
	netinfo[net].arp.next = 0;
	netinfo[net].packet.next = 0;
	netinfo[net].que.next = 0;
	netinfo[net].ip_id = 0x1234;
	netinfo[net].icmp_id = 0x1423;
	netinfo[net].access = NO_ACCESS;
	strncpy((char*)netinfo[net].devname, name, 8);
	return net;
}

int ifdown(int net) {
	int	netip;
	Packet	*cur;

	if(net >= NETNUM) return -1;
	__close(SuperTask, netinfo[net].fd);
	netip = netinfo[net].ip & netinfo[net].mask;
	if((gateway & netinfo[net].mask) == netip) {
		gateway = 0;
	}
	netinfo[net].fd = 0;
	bzero((char*)netinfo[net].devname, 9);
	for(cur = &netinfo[net].arp;cur->next != 0;cur = cur->next) __free(MALLOC_IP, (char*)cur->next);
	for(cur = &netinfo[net].packet;cur->next != 0;cur = cur->next) __free(MALLOC_IP, (char*)cur->next);
	for(cur = &netinfo[net].que;cur->next != 0;cur = cur->next) __free(MALLOC_IP, (char*)cur->next);
	return 0;
}

int set_gateway(int gw) {
	int	net, netip;

	for(net = 0;net < NETNUM;net++) {
		if(netinfo[net].fd != 0) {
			netip = netinfo[net].ip & netinfo[net].mask;
			if((gw & netinfo[net].mask) == netip) {
				gateway = gw;
				gw_net = net;
				return 0;
			}
		}
	}
	return -1;
}

int dns_server(int dns_ip){
	if(dns_ip == 0) return dns;
	dns = dns_ip;
	return 0;
}

int icmp_send(int task_id, int ip, int size) {
	int	net, i, sum, seq;
	char	*buf, c;
	ICMP	*icmp;

	net = get_net(ip);
	if(net == -1) return -1;
	buf = __malloc(SuperTask, sizeof(ICMP) + size);
	if(buf == 0) return -1;
	icmp = (ICMP*)buf;
	icmp->icmp_type = ICMP_ECHO_REQUEST;
	icmp->icmp_code = 0;
	icmp->icmp_cksum = 0;
	icmp->icmp_id = htobe16(task_id);
	seq = (netinfo[net].icmp_id)++;
	icmp->icmp_seq = htobe16(seq);
	c = 'a';
	for(i = 0;i < size;i++) {
		buf[sizeof(ICMP) + i] = c++;
		if(c > 'z') c = 'a';
	}
	sum = checksum(buf, sizeof(ICMP) + size, 0);
	icmp->icmp_cksum = htobe16(sum);
	write_ip(net, buf, sizeof(ICMP) + size, ICMP_PROTOCOL, netinfo[net].ip, ip, 0);
	__free(SuperTask, buf);
	return seq;
}

void ip_interval(void) {
	static		int	count = 0;
	Packet		*packet, *cur, *packed;
	IPHdr		*iphdr;
	int		net, diff, addr, off;
	int		size, id;
	char		*dest, *src;
	unsigned short	sum;

	iphdr = 0;
	off = 0;
	count++;
	udp_interval();
	tcp_interval();
	if(count & 0x03) return;
	for(net = 0;net < NETNUM;net++) {
		if(netinfo[net].fd != 0) {
			for(;;) {
				task_sw_ctl(1);
				packet = 0;
				cur = &netinfo[net].packet;
				while(cur->next != 0) {
					addr = (int)cur->next->data + sizeof(Ether);
					iphdr = (IPHdr*)addr;
					off = be16toh(iphdr->ip_off);
					if((off & IP_MF) == 0) {
						packet = cur->next;
						cur->next = cur->next->next;
						break;
					}
					cur = cur->next;
				}
				task_sw_ctl(0);
				if(packet == 0) break;
				off &= IP_OFFMASK;
				off <<= 3;
				if(off == 0) {
					ip_recv(net, packet);
				} else {
					size = off + be16toh(iphdr->ip_len);
					id = be16toh(iphdr->ip_id);
					packed = (Packet*)__malloc(MALLOC_IP, size + sizeof(IPHdr) + sizeof(Ether) + sizeof(Packet));
					bzero((char*)packed, size + sizeof(IPHdr) + sizeof(Ether) + sizeof(Packet));
					packed->stamp = packet->stamp;
					packed->next = 0;
					dest = (char*)packed->data;
					src = (char*)packet->data;
					memcpy(dest, src, sizeof(IPHdr) + sizeof(Ether));
					dest = &dest[sizeof(IPHdr) + sizeof(Ether)];
					size = be16toh(iphdr->ip_id);
					src = &src[sizeof(IPHdr) + sizeof(Ether)];
					memcpy(&dest[off], src, size);
					__free(MALLOC_IP, (char*)packet);
					addr = (int)packed->data + sizeof(Ether);
					iphdr = (IPHdr*)addr;
					iphdr->ip_off = 0;
					iphdr->ip_sum = 0;
					sum = checksum((char*)iphdr, sizeof(IPHdr), 0);
					iphdr->ip_sum = htobe16(sum);
					task_sw_ctl(1);
					cur = &netinfo[net].packet;
					while(cur->next != 0) {
						addr = (int)cur->next->data + sizeof(Ether);
						iphdr = (IPHdr*)addr;
						if(be16toh(iphdr->ip_id) == id) {
							packet = cur->next;
							src = (char*)packet->data;
							src = &src[sizeof(IPHdr) + sizeof(Ether)];
							size = be16toh(iphdr->ip_len);
							off = be16toh(iphdr->ip_off) & IP_OFFMASK;
							off <<= 3;
							memcpy(&dest[off], src, size);
							cur->next = cur->next->next;
							task_sw_ctl(0);
							__free(MALLOC_IP, (char*)packet);
							task_sw_ctl(1);

						} else {
							cur = cur->next;
						}
					}
					task_sw_ctl(0);
					ip_recv(net, packed);
				}
			}
			if((count & 0x1ff) == 0) {
				for(;;) {
					task_sw_ctl(1);
					packet = 0;
					cur = &netinfo[net].packet;
					while(cur->next != 0) {
						diff = task_time() - cur->next->stamp;
						if(diff > IPDATA_KEEP) {
							packet = cur->next;
							cur->next = cur->next->next;
							break;
						}
						cur = cur->next;
					}
					task_sw_ctl(0);
					if(packet == 0) break;
					__free(MALLOC_IP, (char*)packet);
				}
				for(;;) {
					task_sw_ctl(1);
					packet = 0;
					cur = &netinfo[net].que;
					while(cur->next != 0) {
						diff = task_time() - cur->next->stamp;
						if(diff > IPDATA_KEEP) {
							packet = cur->next;
							cur->next = cur->next->next;
							break;
						}
						cur = cur->next;
					}
					task_sw_ctl(0);
					if(packet == 0) break;
					__free(MALLOC_IP, (char*)packet);
				}
			}
		}
	}
}

void ip_init(void) {
	int	i;
	char	*ptr;

	netinfo = (NetInfo *)__malloc(SuperTask, sizeof(NetInfo) * NETNUM);
	ips = (int*)__malloc(SuperTask, sizeof(int) * CACHESIZ);
	macs = (volatile unsigned char **)__malloc(SuperTask, sizeof(char*) * CACHESIZ);
	ptr = __malloc(SuperTask, CACHESIZ * ETHER_ADDR_LEN);
	netdatbuf = __malloc(SuperTask, ETHER_MAX_LEN);
	arpbuf = __malloc(SuperTask, sizeof(Ether) + sizeof(ARPHdr));
	for(i = 0;i < CACHESIZ;i++) {
		macs[i] = ptr;
		ptr = &ptr[ETHER_ADDR_LEN];
	}
	for(i = 0;i < NETNUM;i++) {
		netinfo[i].fd = 0;
		bzero((char*)netinfo[i].devname, 9);
	}
	for(i = 0;i < CACHESIZ;i++) ips[i] = 0;
	gateway = 0;
	dns = 0;
	udp_init();
	tcp_init();
}
