/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/2/14 first release		*/
/*					*/
/****************************************/
#include "mes.h"
#include "conf.h"
#include <string.h>

#define CELLSIZ	256
static volatile char	*memory;
static volatile short	*table;
static int	tablenum;
#if defined(__arm__)
  static volatile char *fastmem;
  static volatile short fasttbl[MEMSIZ / CELLSIZ];
  static int	FASTADR, FASTSIZ;
#endif
extern unsigned char	_user;
extern BUSConf	memory_config;

char debug_flag;

int mm_init(void) {
	int	i, tablesec;

debug_flag = 0;


#if defined(__arm__)
	FASTADR = (int)&_user;
	FASTSIZ = 0x120000 - FASTADR;
	fastmem = (char*)FASTADR;
	for(i = 0;i < (FASTSIZ / CELLSIZ);i++) fasttbl[i] = 0;
	bzero(fastmem, FASTSIZ);
#endif
	memory = (char*)memory_config.memaddr;
	table = (short*)memory_config.memaddr;
	tablenum = memory_config.memsize / CELLSIZ;
	for(i = 0;i < tablenum;i++) table[i] = 0;
	bzero((char*)memory, memory_config.memsize);
	tablesec = (tablenum * sizeof(short)) / CELLSIZ;
	for(i = 0;i < tablesec;i++) table[i] = SuperTask;
	return memory_config.memaddr + memory_config.memsize;
}

char* __malloc(int id, int size) {
	int	p, i, s, j;
	char	*ptr;

	if(size == 0) return 0;
	i = s = 0;
	for(p = 0;p < tablenum;p++) {
		if(table[p] == 0) {
			i++;
		} else {
			i = 0;
			s = p + 1;
		}
		if(i == ((size - 1) / CELLSIZ + 1)) {
			for(j = 0;j < i;j++) {
				table[s + j] = -1;
			}
			table[s] = id;
			ptr = (char*)&memory[s * CELLSIZ];
			return ptr;
		}
	}
	return 0;
}

int __free(int id, char *ptr) {
	int	p, s;

	s = (int)ptr - (int)memory;
	if(s < 0 || s >= memory_config.memsize) return -1;
	s = s / CELLSIZ;
	if(table[s] != id && id != SuperTask) return -1;
	table[s] = 0;
	for(p = s + 1;;p++) {
		if(table[p] != -1) break;
		table[p] = 0;
	}
	return 0;
}

int change_memid(int id, char *ptr) {
	int	s;

	s = (int)ptr - (int)memory;
	if(s < 0 || s >= memory_config.memsize) return -1;
	s = s / CELLSIZ;
	table[s] = id;
	return 0;
}

char* fast_alloc(int id, int size) {
#if defined(__arm__)
	int	p, i, s, j;

	if(size == 0) return 0;
	i = s = 0;
	for(p = 0;p < (FASTSIZ / CELLSIZ);p++) {
		if(fasttbl[p] == 0) {
			i++;
		} else {
			i = 0;
			s = p + 1;
		}
		if(i == ((size - 1) / CELLSIZ + 1)) {
			for(j = 0;j < i;j++) {
				fasttbl[s + j] = -1;
			}
			fasttbl[s] = id;
			return &fastmem[s * CELLSIZ];
		}
	}
#endif
	return __malloc(id, size);
}

int fast_free(int id, char *ptr) {
#if defined(__arm__)
	int	p, i, s;

	s = (int)ptr - (int)fastmem;
	if(s < 0 || s >= FASTSIZ) return __free(id, ptr);
	s = s / CELLSIZ;
	if(fasttbl[s] != id && id != SuperTask) return -1;
	fasttbl[s] = 0;
	for(p = s + 1;;p++) {
		if(fasttbl[p] != -1) break;
		fasttbl[p] = 0;
	}
	return 0;
#endif
	return __free(id, ptr);
}

int fast_change_memid(int id, char *ptr) {
#if defined(__arm__)
	int	s;

	s = (int)ptr - (int)fastmem;
	if(s < 0 || s >= FASTSIZ) return change_memid(id, ptr);
	s = s / CELLSIZ;
	fasttbl[s] = id;
	return 0;
#endif
	return change_memid(id, ptr);
}

void free_memory(int id) {
	int	i;
	char	flag;
	volatile short *tbl;

	tbl = table;
	flag = 0;
	for(i = 0;i < tablenum;i++) {
		if(*tbl != -1) flag = 0;
		if(*tbl == id) flag = 1;
		if(flag == 1) *tbl = 0;
		tbl++;
	}
#if defined(__arm__)
	for(p = 0;p < (FASTSIZ / CELLSIZ);p++) {
		if(fasttbl[p] != -1) flag = 0;
		if(fatstbl[p] == id) flag = 1;
		if(flag) fasttbl[p] = 0;
	}
#endif
}
