#include "mes2.h"

#define SVC_TASK	0
#define ARP_TASK	1

#define DEVFD_MASK	0x40000000
#define FATFD_MASK	0x20000000
#define FAT_MASK	0xffff

#define MAX_PATH_LEN	512

#define TIMER0_SW	0
#define TIMER1_SW	1
#define TIMER2_SW	2
#define TIMER3_SW	3
#define TIMER4_SW	4

/****************/
/* interrupt.c	*/
/****************/
void int_enable(void);
int int_disable(void);
void int_recover(int reg);
int set_handler(int id, unsigned int number, void (*handler)());
#if defined(__arm__)
  void irq_func(void);
  void core_func(int pc);
#endif
void interrupt_init(void);

/****************/
/* init.c	*/
/****************/
int init_stage1(void);
void init_stage2(void);

/****************/
/* device.c	*/
/****************/
void device_init();
int __open(int id, char *name, int option);
int __close(int id, int fd);
int __write(int id, int fd, char *data, int size);
int __read(int id, int fd, char *data, int size);
int __seek(int id, int fd, int position);
int __ioctl(int id, int fd, int data, int op);
int __flush(int id, int fd);
int add_device(int id, void *func);
int add_device_multi(int id, void *func);
int del_device(int id);
void free_device(int id);
void device_interval();

/****************/
/* pcmcia.c	*/
/****************/
int pccard_init(unsigned int base);

/****************/
/* fat.c	*/
/****************/
int OpenFile(int id, char *_name, int option);
int FlushFile(int id, int fd);
int CloseFile(int id, int fd);
int ReadFile(int id, int fd, char *data, int siz);
int WriteFile(int id, int fd, char *data, int size);
int SeekFile(int id, int fd, int pos);
int GetFileSize(int id, int _fd);
int GetFileId(int _fd);
int SetDirent(int id, char *name, DirEnt *entry);
int NextRecord(int id, int _fd, DirEnt *entry);
void fat_init(void);
int mount(char *devname);
int eject(char *devname);
int format(char *devname);
int free_file(int id);
#define DeleteFile(id, name) __open(id, name, OptRemove);

/****************/
/* memory.c	*/
/****************/
int mm_init(void);
char* __malloc(int id, int size);
int __free(int id, char *ptr);
int change_memid(int id, char *ptr);
char* fast_alloc(int id, int size);
int fast_free(int id, char *ptr);
int fast_change_memid(int id, char *ptr);
void free_memory(int id);

/****************/
/* elf.c	*/
/****************/
int elf_info(int id, int fd, int filesize, int *start, int *size, int *stack, int *got, int *gotsiz);
int elf_load(int id, int fd, int filesize, char *pgm);

/****************/
/* loaddisk.c	*/
/****************/
void load_disk(void);
char *get_romconfig(int *len, char *name);

/****************/
/* task.c	*/
/****************/
int task_time(void);
int random(void);
int cur_task(void);
void task_sw_ctl(int flag);
int set_arg(int _task_id, int cmd, int argc, int data);
int get_arg(int task_id, int cmd, int argc, int *data);
int get_reg(int task_id, int regnum, int *reg);
void system_call(void);
void interval(void);
void set_task_stay(int id);
void wait_ms(int ms);
void task_init(void);
int get_stdio(int id);
int get_stderr(int id);
int execute(int task_id, int argc, char **argv, int stdfd);
char*	__cwd(int task_id);
void task_wakeup(int id, int value);
void init_process(void);

/****************/
/* ip.c		*/
/****************/
short checksum(char *data, int size, int init);
int get_net(int ip);
int get_ipinfo(int net, int *ip, int *mask, char *mac);
int dhcp_server(int net, int serv);
int setip(int net, int ip, int mask);
void send_arp(int net, int ip);
void write_ip(int net, char *data, int size, int proto, int src, int dst, int opt);
void arp_main(void);
int get_netnum(char* name);
int net_ioctl(int net, int data, int op);
int ifconfig(char* name, int ip, int mask);
int ifdown(int net);
int set_gateway(int gw);
int dns_server(int dns_ip);
int icmp_send(int task_id, int ip, int size);
void ip_interval(void);
void ip_init(void);

/****************/
/* udp.c	*/
/****************/
int udp_socket(int task_id);
int udp_free(int task_id, int socket);
int udp_bind(int task_id, int socket, struct sockaddr *bind);
int sendto(int task_id, int socket, char *data, int size, struct sockaddr *target);
void udp_recv(int net, Packet *packet);
int udp_read(int task_id, int socket, char *data, int datsiz, struct sockaddr *target);
void udp_interval(void);
int get_udp_taskid(unsigned short port);
int get_udp_bind(unsigned int sock, struct sockaddr *bind);
int udp_timeout(int task_id, int socket, int ms);
int get_udp_timeout(int task_id, int socket);
void udp_clean(int id);
void udp_init(void);

/****************/
/* tcp.c	*/
/****************/
int check_socket(int sock, int id);
int tcp_close_socket(int id, int sock);
int tcp_reset_socket(int id, int sock);
int tcp_socket(int id, int soctype);
int tcp_bind(int id, int sock, struct sockaddr *bind);
int tcp_synsend(int id, int sock, struct sockaddr *target);
int tcp_synrecv(int id, int sock, struct sockaddr *target);
int tcp_listen(int id, int sock, int max);
int tcp_write(int id, int sock, char *data, int size);
int tcp_get(int id, int sock, char *data, int bufsize);
int tcp_timeout(int id, int sock, int ms);
int get_tcp_timeout(int id, int sock);
void tcp_interval(void);
int tcp_change_taskid(int id, int sock);
void tcp_clean(int id);
void tcp_init(void);

/****************/
/* usb.c	*/
/****************/
int usb_setup(int fd, char *desc, char *data);
int usb_read(int fd, char *data, int size, int ep);
int usb_write(int fd, char *data, int size, int ep);
int usb_get_desc(int num, int *fd, int *class, int *sub);
int usb_get_conf(int num, int *epnum, char *addr, char *attr, char *epsiz);
int usb_product(int num, int *vendor, int *product);
int usb_alloc(int num);
int usb_free(int num);
int usb_handle(int num, int minor, void (*handle)(int));
int usb_mount(int chip);
int usb_allfree(int chip);
int usb_eject(int chip);
int usb_status(int num);
void usb_init(void);
void usb_interval(void);

/****************/
/* shm.c	*/
/****************/
void shm_init(void);
int shm_get(unsigned int key, int size);
char* shm_at(int id);
int shm_dt(int id);

/****************/
/* config.c	*/
/****************/
void config(char *src, char *dst, int size);
void init_register(char *src, char *dst, int size);

/****************/
/* misc.c	*/
/****************/
void write_hex(char *str, int size);
int stoi(char *s);
int stoh(char *s);
void ddr_set(int port, int data);
void ddr_reset(int port, int data);
int shadow_port(int port);
int shadow_bit(int port, int bit);
void ddr_set(int port, int data);
void ddr_reset(int port, int data);
int shadow_port(int port);
int shadow_bit(int port, int bit);
void set_buswidth(int address, int width, int wait);
IOREG* get_ddr(unsigned int num);
IOREG* get_dr(unsigned int num);
void pe_set(int, int);
void pe_reset(int, int);
int hw_config(int, int, int);

/****************/
/* trap.c	*/
/****************/
void task_sw_trap(void);
int command_trap();

/****************/
/* coff.c	*/
/****************/
int stack_size(char *coffarea);
int get_program_size(char *coffarea);
int coff2bin(char* binarea, char* coffarea);

/****************/
/* regist.c	*/
/****************/
int set_regist(int op, char *name, char *str);
int get_regist(int op, char *name, char *str, int maxsize);
int list_regist(int op, int index, char *name);
void init_regist(void);
