#include "mes.h"
#include "id.h"
#include <regdef.h>
#include <ctype.h>

void write_hex(char *str, int size) {
	static char hextbl[] = "0123456789ABCDEF";
	static unsigned int	i, val;
	static unsigned char	buf[2];

	for(i = 0;i < size;i++) {
		val = (int)(str[i]) & 0xff;
		buf[0] = hextbl[(val >> 4) & 0x0f];
		buf[1] = hextbl[val & 0x0f];
		__write(SuperTask, 0, buf, 2);
	}
}

void write_hex2(char *str, int size) {
	static char hextbl[] = "0123456789ABCDEF";
	static unsigned int	i, val;
	static unsigned char	buf[2];

	for(i = 0;i < size;i++) {
		val = (int)(str[i]) & 0xff;
		buf[0] = hextbl[(val >> 4) & 0x0f];
		buf[1] = hextbl[val & 0x0f];
		command_trap(_write_, 0, buf, 2);
	}
}

int stoi(char *s) {
	int	i, n;

	n = 0;
	i = (s[0] == '-') ? 1 : 0;
	for(;isdigit(s[i]);i++) {
		n = 10 * n + (s[i] - '0');
	}
	if(s[0] == '-') n = 0 - n;
	return n;
}

int stoh(char *s) {
	int		i, d;
	unsigned int	n;

	n = 0;
	for(i = 0;;i++) {
		if(isdigit(s[i])) d = s[i] - '0';
		else if(s[i] >= 'a' && s[i] <= 'f') d = s[i] - 'W';
		else break;
		n = (n << 4) + d;
	}
	return n;
}

#if defined(__H8300H__)

static IOREG ddrdata[12];

static IOREG* ddrs[] = {
	0, &P1DDR, &P2DDR, &P3DDR, &P4DDR, &P5DDR, &P6DDR, 0, &P8DDR, &P9DDR, &PADDR, &PBDDR,
};

static IOREG* drs[] = {
	0, &P1DR, &P2DR, &P3DR, &P4DR, &P5DR, &P6DR, &P7DR, &P8DR, &P9DR, &PADR, &PBDR,
};

IOREG* get_dr(unsigned int num) {
	if(num > 11) return 0;
	return drs[num];
}

void ddr_set(int port, int data) {
	ddrdata[port] |= data;
	*(ddrs[port]) = ddrdata[port];
}

void ddr_reset(int port, int data) {
	ddrdata[port] &= ~data;
	*(ddrs[port]) = ddrdata[port];
}

void pe_set(int port, int data) {
}

void pe_reset(int port, int data) {
}

#endif

#if defined(SH_7045)

static IOREG* ddrs[] = {
	&PAIORH, &PAIORL, &PBIOR, &PCIOR, &PDIORH, &PDIORL, &PEIOR,
};

static IOREG* drs[] = {
	&PADRH, &PADRL, &PBDR, &PCDR, &PDDRH, &PDDRL, &PEDR,
};

IOREG* get_dr(unsigned int num) {
	if(num > 6) return 0;
	return drs[num];
}

void ddr_set(int port, int data) {
	*(ddrs[port]) |= data;
}

void ddr_reset(int port, int data) {
	*(ddrs[port]) &= ~data;
}

void pe_set(int port, int data) {
}

void pe_reset(int port, int data) {
}
#endif

#if defined(SH_7706)

static volatile short* ddrs[] = {
	&PACR,&PBCR,&PCCR,&PDCR,&PECR,&PFCR,&PGCR,&PHCR,&SCPCR
};

static IOREG* drs[] = {
	&PADR,&PBDR,&PCDR,&PDDR,&PEDR,&PFDR,&PGDR,&PHDR,&SCPDR,&PJDR
};

IOREG* get_dr(unsigned int num) {
	if(num > 9) return 0;
	return drs[num];
}

void ddr_set(int port, int data) {
	int	i;
	short	reg;

	reg = *(ddrs[port]);
	for(i = 0;i < 8;i++) {
		if(data & (1 << i)) {
			reg &= ~(3 << (i * 2));
			reg |= 1 << (i * 2);
		}
	}
	*(ddrs[port]) = reg;
}

void ddr_reset(int port, int data) {
	int	i;
	short	reg;

	reg = *(ddrs[port]);
	for(i = 0;i < 8;i++) {
		if(data & (1 << i)) {
			reg |= 3 << (i * 2);
		}
	}
	*(ddrs[port]) = reg;
}

void pe_set(int port, int data) {
	ddr_reset(port, data);
}

void pe_reset(int port, int data) {
	int	i;
	short	reg;

	reg = *(ddrs[port]);
	for(i = 0;i < 8;i++) {
		if(data & (1 << i)) {
			reg &= ~(3 << (i * 2));
		}
	}
	*(ddrs[port]) = reg;
}
#endif

void set_buswidth(int address, int width, int wait) {
#if defined(__H8300H__)
	int	i, msk, msk2, inc, adr, bw, ast, w;

	inc = 0x200000;
	adr = 0x200000;
	bw = 0;
	ast = 0;
	msk = 2;
	w = wait;
	if(w > 3) w = 3;
	w <<= 2;
	msk2 = 0xc;
	for(i = 1;i < 8;i++) {
		if(address >= adr && address < (adr + inc)) {
			if(width == 16) {
				bw = msk;
				P4DDR = 0xff;
			}
			if(wait == 0) ast = msk;
			ABWCR &= ~bw;
			ASTCR &= ~ast;
			CSCR |= msk;
  #if defined(H8_3069)
			if(msk2 > 0xff) {
				WCRH &= ~(msk2 >> 8);
				WCRH |= w >> 8;
			} else {
				WCRL &= ~msk2;
				WCRL |= w;
			}
  #endif
			break;
		}
		msk <<= 1;
		msk2 <<= 2;
		w <<= 2;
		adr += inc;
	}
#endif
#if defined(SH_7045)
	int	i, inc, adr, bw, w;

	address &= 0xffffff;
	inc = 0x400000;
	adr = 0x000000;
	bw = 0;
	w = wait;
	if(w > 15) w = 15;
	for(i = 0;i < 4;i++) {
		if(address >= adr && address < (adr + inc)) {
			if(width == 32) {
				PDCRH1 = 0x5555;	/* D31-24 are enable */
				PDCRH2 = 0x5555;	/* D23-16 are enable */
				BCR1 |= 1 << (i + 4);
			} else {
				BCR1 &= ~(1 << (i + 4));
			}
			if(width == 16) BCR1 |= 1 << i;
			if(width == 8) BCR1 &= ~(1 << i);
			WCR1 &= ~(0xf << (i * 4));
			WCR1 |= w << (i * 4);
		}
		adr += inc;
	}
#endif
#if defined(SH_7706)
	int	i, inc, adr, bw, w;

	address &= 0x1fffffff;
	inc = 0x4000000;
	adr = 0x8000000;
	bw = 0;
	w = wait;
	if(w > 7) w = 7;
	for(i = 2;i < 7;i++) {
		if(address >= adr && address < (adr + inc)) {
			switch(width) {
			case 32:
				BCR2 &= ~(3 << (i * 2));
				BCR2 |= 3 << (i * 2);
				break;
			case 16:
				BCR2 &= ~(3 << (i * 2));
				BCR2 |= 2 << (i * 2);
				break;
			case 8:
				BCR2 &= ~(3 << (i * 2));
				BCR2 |= 1 << (i * 2);
				break;
			}
			if(i >= 4) {
				WCR2 &= ~(0x7 << (i * 3 - 5));
				WCR2 |= w << (i * 3 - 5);
			} else if(i >= 2) {
				WCR2 &= ~(0x3 << (i * 2 - 1));
				WCR2 |= w << (i * 3 - 5);
			} else if(i == 0) {
				WCR2 &= ~7;
				WCR2 |= w;
			}
			PCCR &= ~(0x03 << (i * 2 + 2));
		}
		adr += inc;
	}
#endif
}
/*
int hw_config(int op, int data, int option) {
	switch(op) {
#if defined(SH_7706)
	case HW_CONFIG_TURBO:
		value = (data == 0) ? 0x116 : 0x112;
		FRQCR = value;
		asm(
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	nop" "\n\t"
		"	mov	#0x20,r0" "\n\t"
		"	shll8	r0" "\n\t"
		"	dt	r0" "\n\t"
		"	bf	.-2" "\n\t"
		);
		return 0;
#endif
	}
	return -1;
}
*/
