/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2006/6/16 first release		*/
/*					*/
/****************************************/
#include <macro.h>
#include <module.h>
#ifdef _KERNEL_
  #include "mes.h"
  #include "conf.h"
  #define device_main(arg1,arg2) init_rom(void)
  extern BUSConf	romdisk_config[];
  extern BUSConf	flush_config;
#endif

#define	ROM_NUM	2

#define	PL064J		0x0001227e
#define MBM29LV160T	0x000422c4

typedef struct {
	int	top;
	int	start, size;
	int	offset;
	int	romid;
	int	secsize, minisec;
	int	bufaddr;
	int	used;
	char	*buffer;
	char	*topbuf;
} rominfo;
static rominfo		*rom;

static void erase_sec(int minor, int address) {
	volatile short	*word;
	short		prev, cur;
	int		sec;

	word = (volatile short *)rom[minor].start;
	switch(rom[minor].romid) {
	case PL064J:
	case MBM29LV160T:
		sec = (address - rom[minor].start) >> 1;
		word[0x555] = 0xaa;
		word[0x2aa] = 0x55;
		word[0x555] = 0x80;
		word[0x555] = 0xaa;
		word[0x2aa] = 0x55;
		word[sec] = 0x30;
		prev = word[sec] & 0x40;
		cur = word[sec] & 0x40;
		while(prev != cur) {
			prev = cur; 
			cur = word[sec] & 0x40;
		}
		word[0] = 0xf0;
		break;
	}
}

static int write_block(int minor, int address, char *buffer) {
	volatile short	*word, *dst, *src;
	unsigned short	prev, cur;
	unsigned int	addr;
	int		cnt, i;

	word = (volatile short *)rom[minor].start;
	switch(rom[minor].romid) {
	case PL064J:
	case MBM29LV160T:
		cnt = rom[minor].secsize >> 1;
		switch(rom[minor].romid) {
		case PL064J:
			if((address - rom[minor].top) < rom[minor].secsize) cnt = rom[minor].minisec >> 1;
			break;
		case MBM29LV160T:
			if((address - rom[minor].top) >= 0x1fc000) {
				cnt = 0x4000 >> 1;
			} else if((address - rom[minor].top) >= 0x1f8000) {
				cnt = 0x2000 >> 1;
			} else if((address - rom[minor].top) >= 0x1f0000) {
				cnt = 0x8000 >> 1;
			}
			break;
		}
		dst = (volatile short *)address;
		src = (volatile short *)buffer;
		word[0x555] = 0xaa;
		word[0x2aa] = 0x55;
		word[0x555] = 0x20;
		for(i = 0;i < cnt;i++) {
			dst[i] = 0xa0;
			dst[i] = src[i];
			prev = dst[i] & 0x40;
			cur = dst[i] & 0x40;
			while(prev != cur) {
				prev = cur; 
				cur = dst[i] & 0x40;
			}
		}
		word[0] = 0x90;
		word[0] = 0xf0;
		break;
	}
}

static int flush_block(int minor) {
	if(rom[minor].bufaddr != 0) {
		if(rom[minor].romid == MBM29LV160T && (rom[minor].bufaddr - rom[minor].top) >= 0x1f0000) {
			erase_sec(minor, rom[minor].bufaddr);
			write_block(minor, rom[minor].bufaddr, rom[minor].buffer);
			erase_sec(minor, rom[minor].bufaddr + 0x8000);
			write_block(minor, rom[minor].bufaddr, &rom[minor].buffer[0x8000]);
			erase_sec(minor, rom[minor].bufaddr + 0xa000);
			write_block(minor, rom[minor].bufaddr, &rom[minor].buffer[0xa000]);
			erase_sec(minor, rom[minor].bufaddr + 0xc000);
			write_block(minor, rom[minor].bufaddr, &rom[minor].buffer[0xc000]);
		} else {
			erase_sec(minor, rom[minor].bufaddr);
			write_block(minor, rom[minor].bufaddr, rom[minor].buffer);
		}
		rom[minor].bufaddr = 0;
	}
}

static int flush_top(int minor) {
	int	i, m;

	switch(rom[minor].romid) {
	case PL064J:
		for(i = 0, m = 1;i < rom[minor].secsize;i += rom[minor].minisec, m <<= 1) {
			if(rom[minor].used & m) {
				erase_sec(minor, rom[minor].start + i);
				write_block(minor, rom[minor].start + i, rom[minor].topbuf + i);
			}
		}
		rom[minor].used = 0;
		break;
	case MBM29LV160T:
		erase_sec(minor, rom[minor].start);
		write_block(minor, rom[minor].start, rom[minor].topbuf);
		break;
	}
}

static int open_rom(int minor, int option) {
	int	i, ret;
	char	*ptr;

	if(minor >= ROM_NUM) return -1;
	if(rom[minor].romid == 0) return -1;
	ptr = device_malloc(SuperTask, rom[minor].secsize * 2);
	if(ptr == 0) return -1;
	rom[minor].topbuf = ptr;
	rom[minor].buffer = &ptr[rom[minor].secsize];
	rom[minor].bufaddr = 0;
	rom[minor].offset = 0;
	rom[minor].used = 0;
	memcpy(rom[minor].topbuf, (char*)rom[minor].start, rom[minor].secsize);
	return 0;
}

static int close_rom(int minor) {
	if(minor >= ROM_NUM) return -1;
	flush_block(minor);
	flush_top(minor);
	device_free(SuperTask, rom[minor].topbuf);
	return 0;
}

static int write_rom(int minor, char *buffer, int size) {
	int	address, sectop, mask;
	char	*data;

	if(minor >= ROM_NUM) return -1;
	address = rom[minor].offset;
	address += rom[minor].start;
	mask = ~(rom[minor].secsize - 1);
	sectop = address & mask;
	mask = rom[minor].secsize - 1;
	if(sectop == rom[minor].start) {
		rom[minor].used |= 1 << ((address & mask) / rom[minor].minisec);
		memcpy(&rom[minor].topbuf[address & mask], buffer, SECSIZ);
	} else if(sectop == rom[minor].bufaddr) {
		memcpy(&rom[minor].buffer[address & mask], buffer, SECSIZ);
	} else {
		flush_block(minor);
		rom[minor].bufaddr = sectop;
		memcpy(rom[minor].buffer, (char*)sectop, rom[minor].secsize);
		memcpy(&rom[minor].buffer[address & mask], buffer, SECSIZ);
	}
	return size;
}

static int read_rom(int minor, char *buffer, int size) {
	int	address, sectop, mask;
	char	*data;

	if(minor >= ROM_NUM) return -1;
	address = rom[minor].offset;
	address += rom[minor].start;
	mask = ~(rom[minor].secsize - 1);
	sectop = address & mask;
	mask = rom[minor].secsize - 1;
	if(sectop == rom[minor].start) {
		memcpy(buffer, &rom[minor].topbuf[address & mask], SECSIZ);
	} else if(sectop == rom[minor].bufaddr) {
		memcpy(buffer, &rom[minor].buffer[address & mask], SECSIZ);
	} else {
		memcpy(buffer, (char*)address, size);
	}
	return size;
}

static int seek_rom(int minor, int pos) {
	if(minor >= ROM_NUM) return -1;
	if(pos >= rom[minor].size) return -1;
	rom[minor].offset = pos;
	return 0;
}

static int ioctl_rom(int minor, int data, int op) {
	int	ret;

	if(minor >= ROM_NUM) return -1;
	ret = 0;
	switch(op) {
	case DEV_INFO:
		break;
	case DEV_FLUSH:
		flush_block(minor);
		flush_top(minor);
		break;
	case MEDIA_SIZE:
		ret = rom[minor].size;
		break;
	default:
		ret = -1;
	}
	return ret;
}

static Functions func;

int device_main(int argc, char **argv) {
	int		i, id, minor;
	unsigned char	c[2];
	volatile short	*word;

	rom = (rominfo *)malloc(sizeof(rominfo) * ROM_NUM);

#ifndef _KERNEL_
	for(minor = 0;minor < ROM_NUM;minor++) rom[minor].top = -1;
	minor = 0;
	for(i = 1;i < argc;i++) {
		if(memcmp(argv[i], "rom", 3) == 0) {
			minor = stoi(&argv[i][3]);
			if(minor >= ROM_NUM) minor = 0;
		} else if(memcmp(argv[i], "b=", 2) == 0) {
			rom[minor].start = stoh(&argv[i][2]);
		} else if(memcmp(argv[i], "s=", 2) == 0) {
			rom[minor].size = stoh(&argv[i][2]);
		} else if(memcmp(argv[i], "top=", 4) == 0) {
			rom[minor].top = stoh(&argv[i][4]);
		}
	}
#endif
	for(minor = 0;minor < ROM_NUM;minor++) {
#ifdef _KERNEL_
		rom[minor].start = romdisk_config[minor].memaddr;
		rom[minor].size = romdisk_config[minor].memsize;
		rom[minor].top = flush_config.memaddr;
#else
		if(rom[minor].top == -1) rom[minor].top = rom[minor].start;
#endif
		word = (volatile short *)rom[minor].top;
		rom[minor].romid = 0;
		rom[minor].secsize = 0;
		word[0x555] = 0xaa;
		word[0x2aa] = 0x55;
		word[0x555] = 0x90;
		id = (int)word[0] & 0xffff;
		id <<= 16;
		id += (int)word[1] & 0xffff;
		word[0] = 0xf0;
		switch(id) {
		case PL064J:
			rom[minor].romid = id;
			rom[minor].secsize = 0x10000;
			rom[minor].minisec = 0x2000;
			break;
		case MBM29LV160T:
			rom[minor].romid = id;
			rom[minor].secsize = 0x10000;
			break;
		}
	}
	func.open_dev = open_rom;
	func.close_dev = close_rom;
	func.write_dev = write_rom;
	func.read_dev = read_rom;
	func.seek_dev = seek_rom;
	func.ioctl_dev = ioctl_rom;
	func.poll_dev = 0;
	strcpy(&(func.name[0]), "rom");
	device_return(&func);
}
