/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/4/5 First release		*/
/*  2006/7/17 Fixed closing by YAMANEKO	*/
/*					*/
/****************************************/
#include "mes.h"
#include "id.h"
#include "macro.h"
#include "ethernet.h"
#include "ip.h"
#include "tcp.h"

#define TCPNUM		32
#define	DFT_MSS		1460
#define	DFT_TIMOUT	1600

#define SOCK_MASK	0x0000ffff
#define SEQ_MASK	0x00ff0000

#define FLAG_REACK	0x80
#define FLAG_POOL	0x40

typedef volatile struct {
	unsigned int	task_id;
	unsigned int	rem_ip;
	unsigned int	send_unack;
	unsigned int	send_next;
	unsigned int	recv_next;
	unsigned int	time;
	unsigned int	sum;
	unsigned int	parent;
	unsigned int	timeout;
	unsigned int	acklate;
	unsigned int	recv_ack;
	unsigned short	dport;
	unsigned short	sport;
	unsigned char	myflags;
	unsigned char	flags;
	unsigned char	state;
	unsigned char	type;
	unsigned char	retries;
	unsigned char	sockseq;
	char		max_connect;
	Packet2		read;
	Packet2		write;
} TCPInfo;

static TCPInfo		*tcpinfo;
static unsigned short	temp_port;

static int tcp_out(int socket, char *data, int size, int seq, int ack);
static int tcp_mapsocket(IPHdr *iphdr);
static void tcp_sendreset(TCPHdr *tcp, unsigned int remip);
static void tcp_newstate(int socket, int nstate);
static void clear_socket(int socket);

static int tcp_out(int socket, char *data, int size, int seq, int ack) {
	char	*buf, mac[ETHER_ADDR_LEN];
	TCPHdr	*tcp;
	int	net, ip, mask, hsiz, sum, opt;
	Pseudo	*phdr;

	if(socket >= TCPNUM) return -1;
	net = get_net(tcpinfo[socket].rem_ip);
	if(net == -1) return -1;
	hsiz = MIN_TCP_HLEN;
	if(tcpinfo[socket].myflags & SYN_FLAG) hsiz += 8;
	buf = __malloc(SuperTask, sizeof(Pseudo) + hsiz + size);
	if(buf == 0) return -1;
	phdr = (Pseudo*)buf;
	tcp = (TCPHdr*)&buf[sizeof(Pseudo)];
	get_ipinfo(net, &ip, &mask, mac);
	phdr->pseudo_src = htobe32(ip);
	phdr->pseudo_dst = htobe32(tcpinfo[socket].rem_ip);
	phdr->pseudo_pad = 0;
	phdr->pseudo_p = TCP_PROTOCOL;
	phdr->pseudo_len = htobe16(hsiz + size);
	tcp->sport = htobe16(tcpinfo[socket].sport);
	tcp->dport = htobe16(tcpinfo[socket].dport);
	if(seq == 0) {
		tcp->seq = htobe32(tcpinfo[socket].send_unack + tcpinfo[socket].sum);
	} else {
		tcp->seq = htobe32(seq);
	}
	if(ack == 0) {
		tcp->ack = htobe32(tcpinfo[socket].recv_next);
	} else {
		tcp->ack = htobe32(ack);
	}
	tcp->offset = (hsiz << 2) & 0xf0;
	tcp->flags = tcpinfo[socket].myflags & 0x3f; 
	tcp->win = htobe16(TCP_DEF_MTU);
	tcp->cksum = htobe16(0);
	tcp->urgent = htobe16(0);
	if(tcpinfo[socket].myflags & SYN_FLAG) {
		opt = sizeof(Pseudo) + MIN_TCP_HLEN;
		buf[opt++] = TCP_OPTHDR_MSS;
		buf[opt++] = 4;
		buf[opt++] = (char)(DFT_MSS >> 8);
		buf[opt++] = (char)DFT_MSS;
		buf[opt++] = TCP_OPTHDR_NOP;
		buf[opt++] = TCP_OPTHDR_NOP;
		buf[opt++] = TCP_OPTHDR_SACK;
		buf[opt++] = 2;
	}
	if(size > 0) memcpy(&buf[hsiz + sizeof(Pseudo)], data, size);
	sum = checksum(buf, sizeof(Pseudo) + hsiz + size, 0);
	tcp->cksum = htobe16(sum);
	write_ip(net, &buf[sizeof(Pseudo)], hsiz + size, TCP_PROTOCOL, ip, tcpinfo[socket].rem_ip, 0);
	__free(SuperTask, buf);
	return 0;
}

static int tcp_mapsocket(IPHdr *iphdr) {
	int		hsiz, tcpsiz, socket, addr;
	unsigned int	srcip, dstip;
	unsigned short	srcport, dstport;
	unsigned char	flags;
	TCPHdr		*tcp;

	hsiz = (iphdr->ip_ver_hl & 0x0f) << 2;
	srcip = be32toh(iphdr->ip_src);
	dstip = be32toh(iphdr->ip_dst);
	addr = (int)iphdr + hsiz;
	tcp = (TCPHdr*)addr;
	tcpsiz = be16toh(iphdr->ip_len) - hsiz;
	srcport = be16toh(tcp->sport);
	dstport = be16toh(tcp->dport);
	flags = tcp->flags;
	if((flags & ACK_FLAG) == 0 && (flags & SYN_FLAG)) {
		for(socket = 0;socket < TCPNUM;socket++) {
			if(tcpinfo[socket].state == STATE_FREE) continue;
			if(tcpinfo[socket].type != SERVER_TYPE) continue;
			if(tcpinfo[socket].sport != dstport) continue;
			return socket;
		}
	} else {
		for(socket = 0;socket < TCPNUM;socket++) {
			if(tcpinfo[socket].state == STATE_FREE) continue;
			if(tcpinfo[socket].sport != dstport) continue;
			if(tcpinfo[socket].dport != srcport) continue;
			if(tcpinfo[socket].rem_ip != srcip) continue;
			return socket;
		}
	}
	return -1;
}

static void tcp_sendreset(TCPHdr *tcp, unsigned int remip) {
	unsigned short	srcport, dstport, ackno, seqno;
	unsigned char	flags;
	int		socket;

	socket = TCPNUM;
	srcport = be16toh(tcp->sport);
	dstport = be16toh(tcp->dport);
	flags = tcp->flags;
	ackno = be32toh(tcp->ack);
	seqno = be32toh(tcp->seq);
	if(flags & RESET_FLAG) return;
	tcpinfo[socket].rem_ip = remip;
	tcpinfo[socket].dport = srcport;
	tcpinfo[socket].sport = dstport;
	if(flags & ACK_FLAG) {
		tcpinfo[socket].send_unack = ackno;
		tcpinfo[socket].myflags = RESET_FLAG;	
		tcpinfo[socket].recv_next = seqno;
	} else {
		tcpinfo[socket].send_unack = 0;
		tcpinfo[socket].myflags = RESET_FLAG | ACK_FLAG;	
		tcpinfo[socket].recv_next = seqno + 1;
	}
	tcp_out(socket, 0, 0, 0, 0);
}

static void tcp_newstate(int socket, int nstate) {
	tcpinfo[socket].state = nstate;
	tcpinfo[socket].time = task_time();
	switch(tcpinfo[socket].state) {
	case STATE_SYN_SENT:
	case STATE_CLOSING:
	case STATE_FINW1:
		tcpinfo[socket].retries = 3;
		break;
	case STATE_FINW2:
		tcpinfo[socket].retries = 10;
		break;
	case STATE_LAST_ACK:
	case STATE_TIMEDWAIT:
		set_arg(tcpinfo[socket].task_id, _tcp_free, 6, tcpinfo[socket].task_id); 	// YAMANEKO
		set_arg(tcpinfo[socket].task_id, _tcp_close, 6, tcpinfo[socket].task_id); 	// YAMANEKO
		tcpinfo[socket].retries = 1;
		break;
	default:
		tcpinfo[socket].time = 0;
		tcpinfo[socket].retries = 0;
	}
	if(tcpinfo[socket].state != STATE_CONNECTED) tcpinfo[socket].flags = 0;
	return;
}

static void clear_socket(int socket) {
	Packet2	*cur, *packet;
	int	parent;
//	char	seqsv;
	unsigned char seqsv;

	cur = &tcpinfo[socket].read;
	while(cur->next != 0) {
		packet = cur->next;
		cur->next = cur->next->next;
		__free(MALLOC_TCP, (char*)packet);
	}
	cur = &tcpinfo[socket].write;
	while(cur->next != 0) {
		packet = cur->next;
		cur->next = cur->next->next;
		__free(MALLOC_TCP, (char*)packet);
	}
	parent = tcpinfo[socket].parent;
	if(parent != -1) tcpinfo[parent].max_connect++;
	seqsv = tcpinfo[socket].sockseq;
	bzero((char*)&tcpinfo[socket], sizeof(TCPInfo));
	tcpinfo[socket].sockseq = seqsv;
	tcpinfo[socket].state = STATE_FREE;
	tcpinfo[socket].type = NONE_TYPE;
	tcpinfo[socket].timeout = DFT_TIMOUT;
	tcpinfo[socket].parent = -1;
}

int check_socket(int sock, int id) {
	int	socket;
//	char	seq;  // k.fujisawa
	unsigned char	seq; // k.fujisawa

	socket = sock & SOCK_MASK;
	seq = sock >> 16;
	if(socket >= TCPNUM) return -1; // k.fujisawa
	if(tcpinfo[socket].sockseq != seq) return -1;
//	if(socket >= TCPNUM) return -1;  // k.fujisawa
	if(tcpinfo[socket].task_id != id) return -1;
	return socket;
}

int tcp_close_socket(int id, int sock) {
	int	socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	if(tcpinfo[socket].state == STATE_CONNECTED) {
		tcpinfo[socket].myflags = FIN_FLAG;
		tcpinfo[socket].myflags |= ACK_FLAG;
		tcpinfo[socket].acklate = 0;
		tcp_out(socket, 0, 0, 0, 0);
		tcp_newstate(socket, STATE_FINW1);
	}
	return 0;
}

int tcp_reset_socket(int id, int sock) {
	int	socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	if(tcpinfo[socket].state == STATE_CONNECTED) {
		tcpinfo[socket].myflags = RESET_FLAG;
		tcp_out(socket, 0, 0, 0, 0);
	}
	tcp_newstate(socket, STATE_TIMEDWAIT);
	return 0;
}

int tcp_socket(int id, int soctype) {
	int	i, sockseq;

	if(soctype == NONE_TYPE) soctype = CLIENT_TYPE;
	if((soctype != SERVER_TYPE) &&
	   (soctype != CLIENT_TYPE)) return -1;
	for(i = 0;i < TCPNUM;i++) {
		if(tcpinfo[i].state == STATE_FREE) {
			tcpinfo[i].task_id = id;
			tcpinfo[i].state = STATE_RESERVED;
			tcpinfo[i].type = soctype;
			tcpinfo[i].send_next = 0xffffffff;
			tcpinfo[i].sockseq++;
			sockseq = (int)tcpinfo[i].sockseq & 0xff;
			return i | (sockseq << 16);
		}
	
	}
	return -1;
}

int tcp_bind(int id, int sock, struct sockaddr *bind) {
	unsigned short	port;
	int		i, socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	for(i = 0;i < TCPNUM;i++) {
		if(tcpinfo[i].state != STATE_FREE) {
			if(tcpinfo[i].sport == bind->sin_port) return -1;
		}
	}
	tcpinfo[socket].sport = bind->sin_port;
	return 0;
}

int tcp_synsend(int id, int sock, struct sockaddr *target) {
	int	socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	if(tcpinfo[socket].sport == 0) {
		if(++temp_port < 1024) temp_port = 1025;
		tcpinfo[socket].sport = temp_port;
	}
	if( (tcpinfo[socket].type & CLIENT_TYPE) == 0) return -1;
	if((tcpinfo[socket].state != STATE_RESERVED) &&
	   (tcpinfo[socket].state != STATE_LISTENING)) return -1;
	tcpinfo[socket].rem_ip = target->sin_addr;
	tcpinfo[socket].dport = target->sin_port;
	tcpinfo[socket].send_unack = task_time() << 16;
	tcpinfo[socket].send_next = tcpinfo[socket].send_unack + 1;
	tcpinfo[socket].myflags = SYN_FLAG;
	tcp_newstate(socket, STATE_SYN_SENT);
	tcp_out(socket, 0, 0, 0, 0);	// YAMANEKO
	return 0;
}


int tcp_synrecv(int id, int sock, struct sockaddr *target) {
	int	newsock, socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	for(newsock = 0;newsock < TCPNUM;newsock++) {
		if(tcpinfo[newsock].state == STATE_SYN_ACK) {
			tcp_newstate(newsock, STATE_CONNECTED);
			target->sin_addr = tcpinfo[newsock].rem_ip;
			target->sin_port = tcpinfo[newsock].dport;
			return newsock;	
		}
	}
	return -2;
}

int tcp_listen(int id, int sock, int max) {
	int	socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	if(tcpinfo[socket].state != STATE_RESERVED) return -1;
	tcpinfo[socket].type = SERVER_TYPE;
	tcpinfo[socket].max_connect = max;
	tcp_newstate(socket, STATE_LISTENING);
	return 0;
}

int tcp_write(int id, int sock, char *data, int size) {
	Packet2	*cur, *packet;
	int	socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	if(tcpinfo[socket].state != STATE_CONNECTED) return -1;
	if((tcpinfo[socket].send_unack + tcpinfo[socket].sum) != tcpinfo[socket].send_next) {
		return -1;
	}
	packet = (Packet2*)__malloc(MALLOC_TCP, size + sizeof(Packet2));
	if(packet != 0) {
		packet->stamp = task_time();
		packet->seq = tcpinfo[socket].send_unack + tcpinfo[socket].sum;
		packet->retries = TCP_CON_ATTEMPTS;
		packet->size = size;
		packet->next = 0;
		memcpy((char*)packet->data, data, size);
		cur = &tcpinfo[socket].write;
		while(cur->next != 0) cur = cur->next;
		cur->next = packet;
		tcpinfo[socket].myflags = ACK_FLAG | PUSH_FLAG;
		tcp_out(socket, data, size, 0, 0);
		tcpinfo[socket].send_next += size;	
		tcpinfo[socket].sum += size;
		tcpinfo[socket].acklate = 0;
	} else {
		return -1;
	}
	return size;
}

int tcp_get(int id, int sock, char *data, int bufsize) {
	Packet2	*cur, *packet;
	int	ret, size, datsiz, seq;
	char	*rcvdata;
	int	socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	switch(tcpinfo[socket].state) {
	case STATE_CONNECTED:
		cur = &tcpinfo[socket].read;
		if(cur->next == 0) return 0;
		break;
	case STATE_CLOSING:
	case STATE_TIMEDWAIT:
		cur = &tcpinfo[socket].read;
		if(cur->next == 0) {
			if((tcpinfo[socket].myflags & FIN_FLAG) == 0) {
				tcpinfo[socket].recv_next++;
				tcpinfo[socket].myflags = FIN_FLAG | ACK_FLAG;
				tcp_out(socket, 0, 0, 0, 0);
			}
			return -1;
		}
		break;
	default:
		return -1;
	}
	if(cur->next != 0) {
		packet = cur->next;
		if((packet->retries & FLAG_POOL) == 0) {
			cur->next = cur->next->next;
			datsiz = packet->size;
			rcvdata = (char*)packet->data;
			size = (datsiz > bufsize) ? bufsize : datsiz;
			memcpy(data, rcvdata, size);
			__free(MALLOC_TCP, (char*)packet);
			if(tcpinfo[socket].state == STATE_CONNECTED && packet->retries & FIN_FLAG) {	// YAMANEKO
				tcpinfo[socket].recv_next++;
				tcpinfo[socket].myflags = FIN_FLAG | ACK_FLAG;
				tcp_out(socket, 0, 0, 0, 0);
			}
			return size;
		}
	}
	return 0;
}

int tcp_timeout(int id, int sock, int ms) {
	int	i, socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
#if defined(__H8300H__)
	tcpinfo[socket].timeout = (ms == 0) ? DFT_TIMOUT : (ms / 3) << 2;
#endif
#if defined(__sh2__)
	tcpinfo[socket].timeout = (ms == 0) ? DFT_TIMOUT : (ms / 3) << 4;
#endif
#if defined(__sh3__)
	tcpinfo[socket].timeout = (ms == 0) ? DFT_TIMOUT : ms << 1;
#endif
	return 0;
}

int get_tcp_timeout(int id, int sock) {
	int	i, socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	return tcpinfo[socket].timeout;
}

int tcp_recv(int net, IPHdr *iphdr) {
	int		sockseq;
	int		temp, hsiz, thsiz, tcpsiz, ret, dat;
	int		task_id, socket, addr, datasize, newsock;
	char		*rcvdata, flg;
	unsigned int	srcip, dstip, seqno, ackno;
	unsigned short	srcport, dstport;
	unsigned char	flags;
	TCPHdr		*tcp;
	Packet2		*cur, *packet;
	Packet2		*cur2, *packet2;
	struct sockaddr	*target;

	hsiz = (iphdr->ip_ver_hl & 0x0f) << 2;
	srcip = be32toh(iphdr->ip_src);
	dstip = be32toh(iphdr->ip_dst);
	addr = (int)iphdr + hsiz;
	tcp = (TCPHdr*)addr;
	tcpsiz = be16toh(iphdr->ip_len) - hsiz;
	srcport = be16toh(tcp->sport);
	dstport = be16toh(tcp->dport);
	flags = tcp->flags;
	seqno = be32toh(tcp->seq);
	ackno = be32toh(tcp->ack);
	thsiz = (tcp->offset & 0xf0) >> 2;
	datasize = tcpsiz - thsiz;
	addr = (int)tcp + thsiz;
	rcvdata = (char*)addr;
	socket = tcp_mapsocket(iphdr);
	if(socket == -1) {
		if(flags & RESET_FLAG) return -1;
		write_hex((char*)&srcip, 4);
		tcp_sendreset(tcp, srcip);
		return -1;
	}
	task_id = tcpinfo[socket].task_id;
	if(flags & RESET_FLAG) {
		if(tcpinfo[socket].type & SERVER_TYPE) {
			tcp_newstate(socket, STATE_LISTENING);
		} else {
			tcp_newstate(socket, STATE_TIMEDWAIT);
		}
		return -1;
	}
	switch(tcpinfo[socket].state) {
	case STATE_CONNECTED:
	case STATE_SYN_ACK:
	case STATE_TIMEDWAIT:
		if(flags & ACK_FLAG) {
			if(flags & SYN_FLAG) {
				tcpinfo[socket].recv_ack = 0;
				if((seqno + 1) == tcpinfo[socket].recv_next) {
					if(ackno == tcpinfo[socket].send_next) {
						tcpinfo[socket].myflags = ACK_FLAG;
						tcp_out(socket, 0, 0, 0, 0);
						return 0;
					}
				}
				return 0;
			}
			if(tcpinfo[socket].sum > 0) {
				cur2 = &tcpinfo[socket].write;
				while(cur2->next != 0) {
					packet2 = cur2->next;
					if(ackno >= (packet2->seq + packet2->size)) {
						cur2->next = cur2->next->next; 
						tcpinfo[socket].sum -= packet2->size;
						tcpinfo[socket].send_unack += packet2->size;
						__free(MALLOC_TCP, (char*)packet2);
					} else {
						cur2 = cur2->next;
					}
				}
			}
			if(tcpinfo[socket].recv_ack == ackno && datasize == 0) {
				cur2 = &tcpinfo[socket].write;
				while(cur2->next != 0) {
					packet2 = cur2->next;
					if(ackno == packet2->seq) {
						tcpinfo[socket].myflags = ACK_FLAG | PUSH_FLAG;
						tcp_out(socket, (char*)packet2->data, packet2->size, packet2->seq, 0);
					}
					cur2 = cur2->next;
				}
			}
			tcpinfo[socket].recv_ack = ackno;
		}
		if(tcpinfo[socket].recv_next <= seqno) {
			if(datasize > 0) {
				flg = 1;
				cur = &tcpinfo[socket].read;
				while(cur->next != 0) {
					if(cur->next->seq == seqno) flg = 0;
					if(cur->next->seq >= seqno) break;
					cur = cur->next;
				}
				if(flg == 1) {
					packet = (Packet2*)__malloc(MALLOC_TCP, datasize + sizeof(Packet2));
					if(packet != 0) {
						packet->size = datasize;
						packet->stamp = task_time();
						packet->seq = seqno;
						packet->retries = flags;
						if(tcpinfo[socket].recv_next < seqno) packet->retries |= FLAG_POOL;
						memcpy((char*)packet->data, rcvdata, datasize);
						if(tcpinfo[socket].read.next == 0) {
							packet->next = 0;
							tcpinfo[socket].read.next = packet;
						} else {
							packet->next = cur->next;
							cur->next = packet;
						}
					}
				}
			}
			if(tcpinfo[socket].recv_next == seqno) {
				tcpinfo[socket].recv_next = seqno + datasize;
				cur = &tcpinfo[socket].read;
				while(cur->next != 0) {
					cur = cur->next;
					if(tcpinfo[socket].recv_next == cur->seq && cur->retries & FLAG_POOL) {
						tcpinfo[socket].recv_next = cur->seq + cur->size;
						cur->retries &= ~FLAG_POOL;
					}
				}
				tcpinfo[socket].myflags = ACK_FLAG;
				tcpinfo[socket].myflags &= ~FLAG_REACK;
				if(tcpinfo[socket].acklate == 0) tcpinfo[socket].acklate = task_time();
				if(flags & FIN_FLAG) {
					tcpinfo[socket].recv_next = seqno + datasize +1 ;
					tcpinfo[socket].myflags = FIN_FLAG | ACK_FLAG;
					tcp_out(socket, 0, 0, 0, 0);
				}
			} else {
				if((tcpinfo[socket].myflags & FLAG_REACK) == 0) {
					tcpinfo[socket].myflags |= FLAG_REACK;
					tcpinfo[socket].time = task_time();
				}
			}
			if(flags & FIN_FLAG) tcp_newstate(socket, STATE_CLOSING);
		} else {
			tcpinfo[socket].myflags = ACK_FLAG;
			tcp_out(socket, 0, 0, 0, seqno + datasize);
		}
		return 0;
	case STATE_FINW1:
	case STATE_FINW2:
		if(flags & FIN_FLAG) {
			tcpinfo[socket].recv_next = seqno + 1;
			tcpinfo[socket].send_unack++;
			tcpinfo[socket].myflags = ACK_FLAG;
			tcp_out(socket, 0, 0, 0, 0);
		} else {
			if(flags & ACK_FLAG) tcp_newstate(socket, STATE_FINW2);
			return 0;
		}
	case STATE_CLOSING:
		tcp_newstate(socket, STATE_TIMEDWAIT);
		return 0;
	case STATE_SYN_SENT:
		if((flags & SYN_FLAG) && (flags & ACK_FLAG)) {
			if(ackno != tcpinfo[socket].send_next) {
				__write(0,0,"SYN+ACK\r",8);
			} else {
				tcpinfo[socket].recv_next = seqno + 1;
				tcpinfo[socket].send_unack = tcpinfo[socket].send_next;
				tcp_newstate(socket, STATE_CONNECTED);
				tcpinfo[socket].myflags = ACK_FLAG;
				tcp_out(socket, 0, 0, 0, 0);
				set_arg(tcpinfo[socket].task_id, _tcp_connect, 6, tcpinfo[socket].task_id); 
				return 0;
			}
		}
		return -1;
	case STATE_LISTENING:
		if((flags & ACK_FLAG) || !(flags & SYN_FLAG)) __write(0,0,"No SYN\r", 7);
		if(tcpinfo[socket].max_connect-- <= 0) return -1;
		for(newsock = 0;newsock < TCPNUM;newsock++) {
			if(tcpinfo[newsock].state == STATE_FREE) {
				tcpinfo[newsock].parent = socket;
				tcpinfo[newsock].type = CLIENT_TYPE;
				tcpinfo[newsock].task_id = tcpinfo[socket].task_id;
				tcpinfo[newsock].sport = tcpinfo[socket].sport;
				tcpinfo[newsock].rem_ip = srcip;
				tcpinfo[newsock].dport = srcport;
				tcpinfo[newsock].state = STATE_RESERVED;
				tcpinfo[newsock].recv_next = seqno + 1;
				tcpinfo[newsock].send_unack = task_time() << 16;
				tcpinfo[newsock].myflags = SYN_FLAG | ACK_FLAG;
				tcpinfo[newsock].timeout = tcpinfo[socket].timeout;
				tcp_out(newsock, 0, 0, 0, 0);
				tcpinfo[newsock].send_next = tcpinfo[newsock].send_unack + 1;
				tcp_newstate(newsock, STATE_SYN_RECV);
				tcpinfo[socket].dport = 0;
				return 0;
			}
	
		}
		__write(0,0,"No sock\r",8);
		return -1;
	case STATE_SYN_RECV:
		if(flags & ACK_FLAG) {
			if(ackno != tcpinfo[socket].send_next) {
				__write(0,0,"wrong Ack\r",10);
			} else {
				tcpinfo[socket].recv_ack = 0;
				tcpinfo[socket].send_unack = tcpinfo[socket].send_next;
				ret = get_arg(tcpinfo[socket].task_id, _tcp_accept, 1, (int*)&target);
				if(ret == -1) {
					tcp_newstate(socket, STATE_SYN_ACK);
				} else {
					tcpinfo[socket].sockseq++;
					sockseq = (int)tcpinfo[socket].sockseq & 0xff;
					set_arg(tcpinfo[socket].task_id, _tcp_accept, 6, tcpinfo[socket].task_id);
					set_arg(tcpinfo[socket].task_id, _tcp_accept, 5, socket | (sockseq << 16));
					target->sin_addr = tcpinfo[socket].rem_ip;
					target->sin_port = tcpinfo[socket].dport;
					tcp_newstate(socket, STATE_CONNECTED);
				}
				return 0;
			}
		}
		return -1;
	default:
		clear_socket(socket);
		tcp_sendreset(tcp, srcip);
	}
	clear_socket(socket);
	return -1;
}

void tcp_interval(void) {
	int	socket, diff, timeout, s_timout;
	Packet2	*cur, *packet;

	for(socket = 0;socket < TCPNUM;socket++) {
		timeout = tcpinfo[socket].timeout;
		s_timout = DFT_TIMOUT;
		if(tcpinfo[socket].state == STATE_FREE) continue;
		if(tcpinfo[socket].state == STATE_CONNECTED) {
			if(tcpinfo[socket].acklate != 0) {
				diff = task_time() - tcpinfo[socket].acklate;
				if(diff > 100) {
					tcp_out(socket, 0, 0, 0, 0);
					tcpinfo[socket].acklate = 0;
				}
			}
			if(tcpinfo[socket].myflags & FLAG_REACK) {
				diff = task_time() - tcpinfo[socket].time;
				if(diff > s_timout) {
					tcp_out(socket, 0, 0, 0, 0);
					tcpinfo[socket].time = task_time();
				}
			}
			cur = &tcpinfo[socket].write;
			while(cur->next != 0) {
				packet = cur->next;
				diff = task_time() - packet->stamp;
				if(diff > timeout) {
					packet->stamp = task_time();
					if(--(packet->retries) == 0) {
						cur->next = cur->next->next; 
						__free(MALLOC_TCP, (char*)packet);
						tcpinfo[socket].myflags = RESET_FLAG;
						tcp_out(socket, 0, 0, 0, 0);
						clear_socket(socket);
					} else {
						tcpinfo[socket].myflags = ACK_FLAG | PUSH_FLAG;
						tcp_out(socket, (char*)packet->data, packet->size, packet->seq, 0);
						cur = cur->next;
					}
				} else {
					cur = cur->next;
				}
			}
		} else {
			if(tcpinfo[socket].retries == 0) continue;
			diff = task_time() - tcpinfo[socket].time;
			if(diff > timeout) {
				tcpinfo[socket].time = task_time();
				switch(tcpinfo[socket].state) {
				case STATE_SYN_RECV:
					tcpinfo[socket].myflags = RESET_FLAG;
					tcp_out(socket, 0, 0, 0, 0);
					clear_socket(socket);
					break;
				case STATE_SYN_SENT:
				case STATE_CLOSING:
					if(--tcpinfo[socket].retries == 0) {
						tcpinfo[socket].myflags = RESET_FLAG;
						tcp_out(socket, 0, 0, 0, 0);
						clear_socket(socket);
						set_arg(tcpinfo[socket].task_id, _tcp_connect, 6, -1);
					} else {
						tcp_out(socket, 0, 0, 0, 0);
					}
					break;
				case STATE_FINW1:
				case STATE_FINW2:
					__write(0,0,"FIN WAIT retry\r", 15);
					if(--tcpinfo[socket].retries == 0) {
						clear_socket(socket);
						set_arg(tcpinfo[socket].task_id, _tcp_close, 6, tcpinfo[socket].task_id); 
					} else {
						if(tcpinfo[socket].retries < 3) tcp_out(socket, 0, 0, 0, 0);
					}
					break;
				case STATE_TIMEDWAIT:
					if(--tcpinfo[socket].retries == 0) {
						clear_socket(socket);
//						set_arg(tcpinfo[socket].task_id, _tcp_free, 6, tcpinfo[socket].task_id);  // YAMANEKO
//						set_arg(tcpinfo[socket].task_id, _tcp_close, 6, tcpinfo[socket].task_id);  // YAMANEKO
					}
					break;
				}
			}
		}
	}
}

int tcp_status(int id, int sock) {
	int	socket;

	if((socket = check_socket(sock, id)) == -1) return -1;
	if(tcpinfo[socket].state != STATE_CONNECTED) return -1;
	return 0;
}

int tcp_change_taskid(int id, int sock) {
	int	socket;
	char	seq;

	socket = sock & SOCK_MASK;
	seq = sock >> 16;
	if(tcpinfo[socket].sockseq != seq) return -1;
	if(socket >= TCPNUM) return -1;
	tcpinfo[socket].task_id = id;
}

void tcp_clean(int id) {
	int	sock;

	for(sock = 0;sock < TCPNUM;sock++) {
		if(tcpinfo[sock].task_id == id) {
			clear_socket(sock);
		}
	}
}

void tcp_init(void) {
	int	i;

	tcpinfo = (TCPInfo *)__malloc(SuperTask, sizeof(TCPInfo) * (TCPNUM + 1));
	bzero((char*)tcpinfo, sizeof(TCPInfo) * (TCPNUM + 1));
	for(i = 0;i <= TCPNUM;i++) {
		tcpinfo[i].sockseq = 0;
		clear_socket(i);
	}
	temp_port = 1025;
}
