/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/2/14 first release		*/
/*					*/
/****************************************/
#include <macro.h>
#include "mes.h"
#include "id.h"
#include "ethernet.h"
#include "ip.h"
#include "udp.h"

#define UDPNUM	16

typedef volatile struct {
	int		task_id;
	unsigned int	bind_ip;
	unsigned short	bind_port;
	unsigned short	tmp_port;
	int		timeout;
	Packet		recv;
} UDPInfo;

static UDPInfo *udpinfo;
static unsigned short temp_port;

static int port_check(unsigned short port) {
	int		i;

	for(i = 0;i < UDPNUM;i++) {
		if(udpinfo[i].task_id != 0) {
			if(udpinfo[i].bind_port == port) return -1;
			if(udpinfo[i].bind_port == 0 && udpinfo[i].tmp_port == port) return -1;
		}
	}
	return 0;
}

int udp_socket(int task_id) {
	int	i;

	for(i = 0;i < UDPNUM;i++) {
		if(udpinfo[i].task_id == 0) {
			if(++temp_port < 1024) temp_port = 1025;
			while(port_check(temp_port) == -1) temp_port++;
			udpinfo[i].tmp_port = temp_port;
			udpinfo[i].task_id = task_id;
			udpinfo[i].timeout = 3000;
			udpinfo[i].bind_ip = 0;
			udpinfo[i].bind_port = 0;
			udpinfo[i].recv.next = 0;
			return i;
		}
	}
	return -1;	
}

int udp_free(int task_id, int socket) {
	if(socket >= UDPNUM) return -1;
	if(udpinfo[socket].task_id != task_id) return -1;
	udpinfo[socket].task_id = 0;
	return 0;
}

int udp_bind(int task_id, int socket, struct sockaddr *bind) {
	if(socket >= UDPNUM) return -1;
	if(udpinfo[socket].task_id != task_id) return -1;
	if(port_check(bind->sin_port) == -1) return -1;
	udpinfo[socket].bind_ip = bind->sin_addr;
	udpinfo[socket].bind_port = bind->sin_port;
	return 0;
}

int sendto(int task_id, int socket, char *data, int size, struct sockaddr *target) {
	char	*buf, mac[ETHER_ADDR_LEN];
	UDPHdr	*udp;
	int	net, ip, mask;
	unsigned short bport;

	if(socket >= UDPNUM) return -1;
	if(udpinfo[socket].task_id != task_id) return -1;
	net = get_net(target->sin_addr);
	if(net == -1) return -1;
	buf = __malloc(SuperTask, sizeof(UDPHdr) + size);
	udp = (UDPHdr*)buf;
	bport = (udpinfo[socket].bind_port == 0) ? udpinfo[socket].tmp_port : udpinfo[socket].bind_port;
	udp->dg_sport = htobe16(bport);
	udp->dg_dport = htobe16(target->sin_port);
	udp->dg_ulen = htobe16(sizeof(UDPHdr) + size);
	udp->dg_cksum = htobe16(0);
	memcpy(&buf[sizeof(UDPHdr)], data, size);
	get_ipinfo(net, &ip, &mask, mac);
	write_ip(net, buf, sizeof(UDPHdr) + size, UDP_PROTOCOL, ip, target->sin_addr, 0);
	__free(SuperTask, buf);
	return size;
}

void udp_recv(int net, Packet *packet) {
	Packet		*cur;
	IPHdr		*iphdr;
	UDPHdr		*udp;
	char		*buf, *data, *srcdata;
	unsigned char	type;
	unsigned int	sip, dip;
	unsigned short	dport, sport;
	int		i, hsiz, udpsiz, addr, sum, task_id, ret;
	int		datsiz, recvsiz, size;
	struct sockaddr *recvinfo;

	iphdr = (IPHdr*)(packet->data + sizeof(Ether));
	hsiz = (iphdr->ip_ver_hl & 0x0f) << 2;
	sip = be32toh(iphdr->ip_src);
	dip = be32toh(iphdr->ip_dst);
	addr = (int)iphdr + hsiz;
	udp = (UDPHdr*)addr;
	udpsiz = be16toh(iphdr->ip_len) - hsiz;
	recvsiz = be16toh(udp->dg_ulen) - sizeof(UDPHdr);
	dport = be16toh(udp->dg_dport);
	sport = be16toh(udp->dg_sport);
	addr += sizeof(UDPHdr);
	srcdata = (char*)addr;
	for(i = 0;i < UDPNUM;i++) {
		if(udpinfo[i].task_id != 0) {
			if((udpinfo[i].bind_ip == 0 || udpinfo[i].bind_ip == sip) &&
			   (udpinfo[i].bind_port == dport ||
			    (udpinfo[i].bind_port == 0 && udpinfo[i].tmp_port == dport))) {
				task_id = udpinfo[i].task_id;
				ret = get_arg(task_id, _recvfrom, 1, (int*)&data);
				if(ret == -1) {
					packet->next = 0;
					cur = &udpinfo[i].recv;
					while(cur->next != 0) cur = cur->next;
					cur->next = packet;
					return;
				} else {
					get_arg(task_id, _recvfrom, 2, &datsiz);
					get_arg(task_id, _recvfrom, 3, (int*)&recvinfo);
					size = (datsiz < recvsiz) ? datsiz : recvsiz;
					memcpy(data, srcdata, size);
					recvinfo->sin_addr = sip;
					recvinfo->sin_port = sport;
					set_arg(task_id, _recvfrom, 6, dport);
					set_arg(task_id, _recvfrom, 5, size);
					break;
				}
			}
		}
	}
	__free(MALLOC_IP, (char*)packet);
}

int udp_read(int task_id, int socket, char *data, int datsiz, struct sockaddr *target) {
	Packet		*cur, *packet;
	IPHdr		*iphdr;
	UDPHdr		*udp;
	unsigned int	sip, dip;
	unsigned short	dport, sport;
	int		addr, size, recvsiz, hsiz;
	char		*srcdata;

	cur = &udpinfo[socket].recv;
	if(cur->next != 0) {
		packet = cur->next;
		cur->next = cur->next->next;
		iphdr = (IPHdr*)(packet->data + sizeof(Ether));
		hsiz = (iphdr->ip_ver_hl & 0x0f) << 2;
		sip = be32toh(iphdr->ip_src);
		dip = be32toh(iphdr->ip_dst);
		addr = (int)iphdr + hsiz;
		udp = (UDPHdr*)addr;
		recvsiz = be16toh(udp->dg_ulen) - sizeof(UDPHdr);
		dport = be16toh(udp->dg_dport);
		sport = be16toh(udp->dg_sport);
		addr += sizeof(UDPHdr);
		srcdata = (char*)addr;
		size = (datsiz < recvsiz) ? datsiz : recvsiz;
		memcpy(data, srcdata, size);
		target->sin_addr = sip;
		target->sin_port = sport;
		set_arg(task_id, _recvfrom, 6, dport);
		set_arg(task_id, _recvfrom, 5, size);
		__free(MALLOC_IP, (char*)packet);
		return size;
	}
	return -1;
}

void udp_interval(void) {
	Packet	*cur, *packet;
	int	diff, i;

	for(i = 0;i < UDPNUM;i++) {
		if(udpinfo[i].task_id != 0) {
			cur = &udpinfo[i].recv;
			while(cur->next != 0) {
				packet = cur->next;
				diff = task_time() - packet->stamp;
				if(diff > udpinfo[i].timeout * 2) {
					cur->next = cur->next->next;
					__free(MALLOC_IP, (char*)packet);
				} else {
					cur = cur->next;
				}
			}
		}
	}
}

int get_udp_taskid(unsigned short port) {
	int	i;

	for(i = 0;i < UDPNUM;i++) {
		if(udpinfo[i].bind_port == port) return udpinfo[i].task_id;
	}
	return -1;
}

int get_udp_bind(unsigned int sock, struct sockaddr *bind) {
	if(sock >= UDPNUM) return -1;
	bind->sin_addr = udpinfo[sock].bind_ip;	
	bind->sin_port = udpinfo[sock].bind_port;
	return udpinfo[sock].task_id;
}

int udp_timeout(int task_id, int socket, int ms) {
	int	i;

	if(socket >= UDPNUM) return -1;
	if(udpinfo[socket].task_id != task_id) return -1;
#if defined(__H8300H__)
	udpinfo[socket].timeout = (ms / 3) << 2;
#endif
#if defined(__sh2__)
	udpinfo[socket].timeout = (ms / 3) << 4;
#endif
#if defined(__sh3__)
	udpinfo[socket].timeout = ms << 1;
#endif
	return 0;
}

int get_udp_timeout(int task_id, int socket) {
	int	i;

	if(socket >= UDPNUM) return -1;
	if(udpinfo[socket].task_id != task_id) return -1;
	return udpinfo[socket].timeout;
}

void udp_clean(int id) {
	int	socket;

	for(socket = 0;socket < UDPNUM;socket++) udp_free(id, socket);
}

void udp_init(void) {
	int	i;

	udpinfo = (UDPInfo *)__malloc(SuperTask, sizeof(UDPInfo) * UDPNUM);
	for(i = 0;i < UDPNUM;i++) {
		udpinfo[i].task_id = 0;
	}
	temp_port = 1025;
}
